@@ -101,18 +101,29 @@ static const char * const sched_node_type_str[] = {
 	[SCHED_NODE_TYPE_VPORTS_TC_TSAR] = "vports TC TSAR",
 };
 
+static void esw_qos_node_attach_to_parent(struct mlx5_esw_sched_node *node)
+{
+	if (!node->parent)
+		list_add_tail(&node->entry, &node->esw->qos.domain->nodes);
+	else
+		list_add_tail(&node->entry, &node->parent->children);
+}
+
 static void
 esw_qos_node_set_parent(struct mlx5_esw_sched_node *node, struct mlx5_esw_sched_node *parent)
 {
-	list_del_init(&node->entry);
 	if (node->type != SCHED_NODE_TYPE_VPORTS_TC_TSAR) {
 		if (node->parent)
 			node->parent->num_vports--;
-		parent->num_vports++;
+		/* node0 is removed, parent could be NULL */
+		if (parent)
+			parent->num_vports++;
 	}
+	list_del_init(&node->entry);
 	node->parent = parent;
-	list_add_tail(&node->entry, &parent->children);
-	node->esw = parent->esw;
+	if (parent)
+		node->esw = parent->esw;
+        esw_qos_node_attach_to_parent(node);
 }
 
 static void
@@ -448,6 +459,7 @@ static int esw_qos_create_node_sched_elem(struct mlx5_core_dev *dev, u32 parent_
 static int esw_qos_vport_create_sched_element(struct mlx5_esw_sched_node *vport_node,
 					      struct netlink_ext_ack *extack)
 {
+	struct mlx5_esw_sched_node *parent = vport_node->parent;
 	u32 sched_ctx[MLX5_ST_SZ_DW(scheduling_context)] = {};
 	struct mlx5_core_dev *dev = vport_node->esw->dev;
 	struct mlx5_vport *vport = vport_node->vport;
@@ -468,7 +480,8 @@ static int esw_qos_vport_create_sched_element(struct mlx5_esw_sched_node *vport_
 		MLX5_SET(vport_element, attr, eswitch_owner_vhca_id,
 			 MLX5_CAP_GEN(vport->dev, vhca_id));
 	}
-	MLX5_SET(scheduling_context, sched_ctx, parent_element_id, vport_node->parent->ix);
+	MLX5_SET(scheduling_context, sched_ctx, parent_element_id,
+		 parent ? parent->ix : vport_node->esw->qos.root_tsar_ix);
 	MLX5_SET(scheduling_context, sched_ctx, max_average_bw, vport_node->max_rate);
 
 	return esw_qos_node_create_sched_element(vport_node, sched_ctx, extack);
@@ -489,8 +502,6 @@ static void esw_qos_destroy_sysfs_rate_group(struct mlx5_eswitch *esw,
 {
 	if (!group)
 		return;
-	if (group == esw->qos.node0)
-		return;
 
 	if (vport->vport != MLX5_VPORT_PF)
 		group->num_vports--;
@@ -776,43 +787,15 @@ static int esw_qos_create(struct mlx5_eswitch *esw, struct netlink_ext_ack *exta
 		return err;
 	}
 
-	if (MLX5_CAP_QOS(dev, log_esw_max_sched_depth)) {
-		esw->qos.node0 = __esw_qos_create_vports_sched_node(esw, 0, NULL, extack);
-	} else {
-		/* The eswitch doesn't support scheduling nodes.
-		 * Create a software-only node0 using the root TSAR to attach vport QoS to.
-		 */
-		if (!__esw_qos_alloc_node(esw, esw->qos.root_tsar_ix,
-					  SCHED_NODE_TYPE_VPORTS_TSAR, NULL))
-			esw->qos.node0 = ERR_PTR(-ENOMEM);
-	}
-	if (IS_ERR(esw->qos.node0)) {
-		err = PTR_ERR(esw->qos.node0);
-		esw_warn(dev, "E-Switch create rate node 0 failed (%d)\n", err);
-		goto err_node0;
-	}
 	refcount_set(&esw->qos.refcnt, 1);
 
 	return 0;
-
-err_node0:
-	if (mlx5_destroy_scheduling_element_cmd(esw->dev, SCHEDULING_HIERARCHY_E_SWITCH,
-						esw->qos.root_tsar_ix))
-		esw_warn(esw->dev, "E-Switch destroy root TSAR failed.\n");
-
-	return err;
 }
 
 static void esw_qos_destroy(struct mlx5_eswitch *esw)
 {
 	int err;
 
-	if (esw->qos.node0->ix != esw->qos.root_tsar_ix)
-		__esw_qos_destroy_node(esw->qos.node0, NULL);
-	else
-		__esw_qos_free_node(esw->qos.node0);
-	esw->qos.node0 = NULL;
-
  	err = mlx5_destroy_scheduling_element_cmd(esw->dev,
 						  SCHEDULING_HIERARCHY_E_SWITCH,
 						  esw->qos.root_tsar_ix);
@@ -1004,7 +987,7 @@ static int esw_qos_vport_tc_enable(struct mlx5_vport *vport, enum sched_node_typ
 err_sched_nodes:
 	if (type == SCHED_NODE_TYPE_RATE_LIMITER) {
 		esw_qos_node_destroy_sched_element(vport_node, NULL);
-		list_add_tail(&vport_node->entry, &vport_node->parent->children);
+		esw_qos_node_attach_to_parent(vport_node);
 	} else {
 		esw_qos_tc_arbiter_scheduling_teardown(vport_node, NULL);
 	}
@@ -1168,7 +1151,6 @@ out:
 static void esw_qos_vport_disable(struct mlx5_vport *vport, struct netlink_ext_ack *extack)
 {
 	struct mlx5_esw_sched_node *vport_node = vport->qos.sched_node;
-	struct mlx5_esw_sched_node *parent = vport_node->parent;
 	enum sched_node_type curr_type = vport_node->type;
 
 	if (curr_type == SCHED_NODE_TYPE_VPORT)
@@ -1178,7 +1160,7 @@ static void esw_qos_vport_disable(struct mlx5_vport *vport, struct netlink_ext_a
 
 	vport_node->bw_share = 0;
 	list_del_init(&vport_node->entry);
-	esw_qos_normalize_min_rate(parent->esw, parent, extack);
+	esw_qos_normalize_min_rate(vport_node->esw, vport_node->parent, extack);
 
 	trace_mlx5_esw_vport_qos_destroy(vport_node->esw->dev, vport);
 }
@@ -1187,22 +1169,23 @@ static int esw_qos_vport_enable(struct mlx5_vport *vport, enum sched_node_type t
 				struct mlx5_esw_sched_node *parent,
 				struct netlink_ext_ack *extack)
 {
+	struct mlx5_esw_sched_node *vport_node = vport->qos.sched_node;
 	int err;
 
 	esw_assert_qos_lock_held(vport->dev->priv.eswitch);
 
-	esw_qos_node_set_parent(vport->qos.sched_node, parent);
+	esw_qos_node_set_parent(vport_node, parent);
 	if (type == SCHED_NODE_TYPE_VPORT)
-		err = esw_qos_vport_create_sched_element(vport->qos.sched_node, extack);
+		err = esw_qos_vport_create_sched_element(vport_node, extack);
 	else
 		err = esw_qos_vport_tc_enable(vport, type, extack);
 	if (err)
 	       return err;
 
-	vport->qos.sched_node->type = type;
-	esw_qos_normalize_min_rate(parent->esw, parent, extack);
-	trace_mlx5_esw_vport_qos_create(vport->dev, vport, vport->qos.sched_node->max_rate,
-					vport->qos.sched_node->bw_share);
+	vport_node->type = type;
+	esw_qos_normalize_min_rate(vport_node->esw, parent, extack);
+	trace_mlx5_esw_vport_qos_create(vport->dev, vport, vport_node->max_rate,
+					vport_node->bw_share);
 
 	return 0;
 }
@@ -1213,6 +1196,7 @@ int mlx5_esw_qos_vport_enable(struct mlx5_vport *vport, enum sched_node_type typ
 {
 	struct mlx5_eswitch *esw = vport->dev->priv.eswitch;
 	struct mlx5_esw_sched_node *sched_node;
+	struct mlx5_eswitch *parent_esw;
 	int err;
 
 	esw_assert_qos_lock_held(esw);
@@ -1223,16 +1207,19 @@ int mlx5_esw_qos_vport_enable(struct mlx5_vport *vport, enum sched_node_type typ
 	if (err)
 		return err;
 
-	parent = parent ?: esw->qos.node0;
-	sched_node = __esw_qos_alloc_node(parent->esw, 0, type, parent);
+	parent_esw = parent ? parent->esw : esw;
+	sched_node = __esw_qos_alloc_node(parent_esw, 0, type, parent);
 	if (!sched_node)
 		return -ENOMEM;
+	if (!parent)
+		list_add_tail(&sched_node->entry, &esw->qos.domain->nodes);
 
 	sched_node->max_rate = max_rate;
 	sched_node->min_rate = min_rate;
 	sched_node->vport = vport;
 	vport->qos.sched_node = sched_node;
-	parent->num_vports++;
+	if (parent)
+		parent->num_vports++;
 
 	err = esw_qos_vport_enable(vport, type, parent, extack);
 	if (err) {
@@ -1285,7 +1272,7 @@ void mlx5_esw_qos_vport_disable(struct mlx5_vport *vport)
 		goto unlock;
 
 	parent = vport->qos.sched_node->parent;
-	WARN(parent != esw->qos.node0, "Disabling QoS on port before detaching it from node");
+	WARN(parent, "Disabling QoS on port before detaching it from node");
 
 	esw_qos_vport_disable(vport, NULL);
 	mlx5_esw_qos_vport_qos_free(vport);
@@ -1382,8 +1369,10 @@ static bool esw_qos_validate_unsupported_tc_bw(struct mlx5_eswitch *esw, u32 *tc
 
 static bool esw_qos_vport_validate_unsupported_tc_bw(struct mlx5_vport *vport, u32 *tc_bw)
 {
-	struct mlx5_eswitch *esw = vport->qos.sched_node ?
-				   vport->qos.sched_node->parent->esw : vport->dev->priv.eswitch;
+	struct mlx5_esw_sched_node *node = vport->qos.sched_node;
+	struct mlx5_eswitch *esw = vport->dev->priv.eswitch;
+
+	esw = (node && node->parent) ? node->parent->esw : esw;
 
 	return esw_qos_validate_unsupported_tc_bw(esw, tc_bw);
 }
@@ -1398,7 +1387,6 @@ static int esw_qos_vport_update(struct mlx5_vport *vport, enum sched_node_type t
 	int err;
 
 	esw_assert_qos_lock_held(vport->dev->priv.eswitch);
-	parent = parent ?: curr_parent;
 	if (curr_type == type && curr_parent == parent)
 		return 0;
 
@@ -1407,8 +1395,10 @@ static int esw_qos_vport_update(struct mlx5_vport *vport, enum sched_node_type t
 		return err;
 
 	if (curr_type == SCHED_NODE_TYPE_TC_ARBITER_TSAR && curr_type == type) {
+		struct mlx5_eswitch *esw = parent ? parent->esw : vport->dev->priv.eswitch;
+
 		esw_qos_tc_arbiter_get_bw_shares(vport->qos.sched_node, curr_tc_bw);
-		if (!esw_qos_validate_unsupported_tc_bw(parent->esw, curr_tc_bw)) {
+		if (!esw_qos_validate_unsupported_tc_bw(esw, curr_tc_bw)) {
 			NL_SET_ERR_MSG_MOD(extack,
 					   "Unsupported traffic classes on the new device");
 			return -EOPNOTSUPP;
@@ -1438,16 +1428,16 @@ int esw_qos_vport_update_parent(struct mlx5_vport *vport, struct mlx5_esw_sched_
 
 	esw_assert_qos_lock_held(esw);
 	curr_parent = vport->qos.sched_node->parent;
-	parent = parent ?: esw->qos.node0;
 	if (curr_parent == parent)
 		return 0;
 
 	/* Set vport QoS type based on parent node type if different from default QoS;
 	 * otherwise, use the vport's current QoS type.
 	 */
-	if (parent->type == SCHED_NODE_TYPE_TC_ARBITER_TSAR)
+	if (parent && parent->type == SCHED_NODE_TYPE_TC_ARBITER_TSAR)
 		type = SCHED_NODE_TYPE_RATE_LIMITER;
-	else if (curr_parent->type == SCHED_NODE_TYPE_TC_ARBITER_TSAR)
+	else if (curr_parent &&
+		 curr_parent->type == SCHED_NODE_TYPE_TC_ARBITER_TSAR)
 		type = SCHED_NODE_TYPE_VPORT;
 	else
 		type = vport->qos.sched_node->type;
@@ -1804,7 +1794,8 @@ static int rate_leaf_tc_bw_set(void *priv, u32 *tc_bw, struct netlink_ext_ack *e
 
 	if (disable) {
 		if (vport_node->type == SCHED_NODE_TYPE_TC_ARBITER_TSAR)
-			err = esw_qos_vport_update(vport, SCHED_NODE_TYPE_VPORT, NULL, extack);
+			err = esw_qos_vport_update(vport, SCHED_NODE_TYPE_VPORT,
+						   vport_node->parent, extack);
 		goto unlock;
 	}
 
@@ -1813,7 +1804,8 @@ static int rate_leaf_tc_bw_set(void *priv, u32 *tc_bw, struct netlink_ext_ack *e
 						extack);
 		vport_node = vport->qos.sched_node;
 	} else {
-		err = esw_qos_vport_update(vport, SCHED_NODE_TYPE_TC_ARBITER_TSAR, NULL, extack);
+		err = esw_qos_vport_update(vport, SCHED_NODE_TYPE_TC_ARBITER_TSAR,
+					   vport_node->parent, extack);
 	}
 	if (!err)
 		esw_qos_set_tc_arbiter_bw_shares(vport_node, tc_bw, extack);
@@ -2076,7 +2068,7 @@ int mlx5_esw_qos_vport_update_parent(struct mlx5_vport *vport, struct mlx5_esw_s
 void
 mlx5_esw_qos_pre_cleanup(struct mlx5_core_dev *dev, int num_vfs)
 {
-	struct mlx5_esw_sched_node *group, *tmp1, *tmp2, *node0, *vport_group;
+	struct mlx5_esw_sched_node *group, *tmp1, *tmp2, *vport_group;
 	struct mlx5_eswitch *esw = dev->priv.eswitch;
 	struct mlx5_vport *vport;
 	int err;
@@ -2115,11 +2107,11 @@ mlx5_esw_qos_pre_cleanup(struct mlx5_core_dev *dev, int num_vfs)
 			if (vport->dev != dev && group->esw->dev != dev)
 				continue;
 
-			node0 = vport->dev->priv.eswitch->qos.node0;
-			err = esw_qos_vport_update_parent(vport, node0, NULL);
+			err = esw_qos_vport_update_parent(vport, NULL, NULL);
 			if (err)
 				esw_warn(vport->dev,
-					 "failed to move vport %d to node0\n", vport->vport);
+					 "failed to move vport %d to root node\n",
+					 vport->vport);
 		}
 		if (!group->num_vports) {
 			esw = group->esw;
