@@ -20,7 +20,7 @@
 #include <linux/string.h>
 #include <linux/wait.h>
 #include <linux/inet.h>
-#include <asm/unaligned.h>
+#include <linux/unaligned.h>
 
 #include <rdma/ib_verbs.h>
 #include <rdma/rdma_cm.h>
@@ -42,6 +42,8 @@
 #define NVMET_RDMA_MAX_MDTS			8
 #define NVMET_RDMA_MAX_METADATA_MDTS		5
 
+#define NVMET_RDMA_BACKLOG 128
+
 struct nvmet_rdma_srq;
 
 struct nvmet_rdma_cmd {
@@ -56,7 +58,6 @@ struct nvmet_rdma_cmd {
 
 enum {
 	NVMET_RDMA_REQ_INLINE_DATA	= (1 << 0),
-	NVMET_RDMA_REQ_INVALIDATE_RKEY	= (1 << 1),
 };
 
 struct nvmet_rdma_rsp {
@@ -107,6 +108,7 @@ struct nvmet_rdma_queue {
 	struct nvmet_rdma_cmd	*cmds;
 
 	struct work_struct	release_work;
+	struct work_struct	release_offload_work;
 	struct work_struct	disconnect_work;
 	struct list_head	rsp_wait_list;
 	struct list_head	rsp_wr_wait_list;
@@ -122,6 +124,8 @@ struct nvmet_rdma_queue {
 
 	bool                    offload;
 	struct nvmet_rdma_xrq   *xrq;
+	bool			offload_free_started;
+	struct mutex		offload_release_lock;
 };
 
 struct nvmet_rdma_port {
@@ -532,12 +536,8 @@ nvmet_rdma_alloc_rsps(struct nvmet_rdma_queue *queue)
 	return 0;
 
 out_free:
-	while (--i >= 0) {
-		struct nvmet_rdma_rsp *rsp = &queue->rsps[i];
-
-		list_del(&rsp->free_list);
-		nvmet_rdma_free_rsp(ndev, rsp);
-	}
+	while (--i >= 0)
+		nvmet_rdma_free_rsp(ndev, &queue->rsps[i]);
 	kfree(queue->rsps);
 out:
 	return ret;
@@ -548,12 +548,8 @@ static void nvmet_rdma_free_rsps(struct nvmet_rdma_queue *queue)
 	struct nvmet_rdma_device *ndev = queue->dev;
 	int i, nr_rsps = queue->recv_queue_size * 2;
 
-	for (i = 0; i < nr_rsps; i++) {
-		struct nvmet_rdma_rsp *rsp = &queue->rsps[i];
-
-		list_del(&rsp->free_list);
-		nvmet_rdma_free_rsp(ndev, rsp);
-	}
+	for (i = 0; i < nr_rsps; i++)
+		nvmet_rdma_free_rsp(ndev, &queue->rsps[i]);
 	kfree(queue->rsps);
 }
 
@@ -779,7 +775,7 @@ static void nvmet_rdma_queue_response(struct nvmet_req *req)
 	struct rdma_cm_id *cm_id = rsp->queue->cm_id;
 	struct ib_send_wr *first_wr;
 
-	if (rsp->flags & NVMET_RDMA_REQ_INVALIDATE_RKEY) {
+	if (rsp->invalidate_rkey) {
 		rsp->send_wr.opcode = IB_WR_SEND_WITH_INV;
 		rsp->send_wr.ex.invalidate_rkey = rsp->invalidate_rkey;
 	} else {
@@ -962,10 +958,8 @@ static u16 nvmet_rdma_map_sgl_keyed(struct nvmet_rdma_rsp *rsp,
 		goto error_out;
 	rsp->n_rdma += ret;
 
-	if (invalidate) {
+	if (invalidate)
 		rsp->invalidate_rkey = key;
-		rsp->flags |= NVMET_RDMA_REQ_INVALIDATE_RKEY;
-	}
 
 	return 0;
 
@@ -1104,6 +1098,7 @@ static void nvmet_rdma_recv_done(struct ib_cq *cq, struct ib_wc *wc)
 	rsp->req.cmd = cmd->nvme_cmd;
 	rsp->req.port = queue->port;
 	rsp->n_rdma = 0;
+	rsp->invalidate_rkey = 0;
 
 	if (unlikely(queue->state != NVMET_RDMA_Q_LIVE)) {
 		unsigned long flags;
@@ -1362,7 +1357,7 @@ static int nvmet_rdma_create_queue_ib(struct nvmet_rdma_queue *queue)
 	qp_attr.cap.max_send_wr = queue->send_queue_size + 1;
 	factor = rdma_rw_mr_factor(ndev->device, queue->cm_id->port_num,
 				   1 << NVMET_RDMA_MAX_MDTS);
-	qp_attr.cap.max_rdma_ctxs = queue->send_queue_size * factor;
+	qp_attr.cap.max_rdma_ctxs = (queue->offload ? 1 : queue->send_queue_size) * factor;
 	qp_attr.cap.max_send_sge = max(ndev->device->attrs.max_sge_rd,
 					ndev->device->attrs.max_send_sge);
 
@@ -1417,12 +1412,27 @@ err_destroy_xrq:
 	goto out;
 }
 
-static void nvmet_rdma_destroy_queue_ib(struct nvmet_rdma_queue *queue)
+static void nvmet_rdma_free_queue_start(struct nvmet_rdma_queue *queue)
 {
-	/* todo: bug #974802 - need to debug with fw */
+	pr_debug("start freeing offload queue %d\n", queue->idx);
+
+	queue->offload_free_started = true;
+	nvmet_sq_destroy(&queue->nvme_sq);
+
 	ib_drain_qp(queue->qp);
 	if (queue->cm_id)
 		rdma_destroy_id(queue->cm_id);
+}
+
+static void nvmet_rdma_destroy_queue_ib(struct nvmet_rdma_queue *queue,
+					bool drain_qp)
+{
+	if (drain_qp) {
+		ib_drain_qp(queue->qp);
+		if (queue->cm_id)
+			rdma_destroy_id(queue->cm_id);
+	}
+
 	ib_destroy_qp(queue->qp);
 	ib_cq_pool_put(queue->cq, queue->recv_queue_size + 2 *
 		       queue->send_queue_size + 1);
@@ -1430,32 +1440,73 @@ static void nvmet_rdma_destroy_queue_ib(struct nvmet_rdma_queue *queue)
 		kref_put(&queue->xrq->ref, nvmet_rdma_destroy_xrq);
 }
 
-static void nvmet_rdma_free_queue(struct nvmet_rdma_queue *queue)
+static void nvmet_rdma_free_queue(struct nvmet_rdma_queue *queue,
+				  bool drain_queue)
 {
-	pr_debug("freeing queue %d\n", queue->idx);
+	pr_debug("freeing queue %d drain_queue %d\n", queue->idx, drain_queue);
 
-	nvmet_sq_destroy(&queue->nvme_sq);
+	if (drain_queue)
+		nvmet_sq_destroy(&queue->nvme_sq);
 
-	nvmet_rdma_destroy_queue_ib(queue);
+	nvmet_rdma_destroy_queue_ib(queue, drain_queue);
 	if (!queue->nsrq && !queue->offload) {
 		nvmet_rdma_free_cmds(queue->dev, queue->cmds,
 				queue->recv_queue_size,
 				!queue->host_qid);
 	}
 	nvmet_rdma_free_rsps(queue);
+	cancel_work_sync(&queue->disconnect_work);
 	ida_free(&nvmet_rdma_queue_ida, queue->idx);
 	kfree(queue);
 }
 
+static void nvmet_rdma_release_queue(struct nvmet_rdma_queue *queue)
+{
+	struct nvmet_rdma_device *dev = queue->dev;
+
+	if (queue->offload &&
+	    dev->device->attrs.nvmf_caps.qp_last_nvme_cqe_reached) {
+		/*
+		 * For offload queues with last NVMe CQE event support, this
+		 * function is called twice to handle potential delays in event
+		 * arrival (e.g., with slow SSDs).
+		 *
+		 * First call: Initiates the queue free process.
+		 * Second call: Completes the queue free process.
+		 *
+		 * The queue is fully freed when receiving the last NVMe CQE
+		 * event for this queue.
+		 * Note: The order of these calls and the event arrival is not
+		 * guaranteed.
+		 */
+		mutex_lock(&queue->offload_release_lock);
+		if (!queue->offload_free_started) {
+			nvmet_rdma_free_queue_start(queue);
+			mutex_unlock(&queue->offload_release_lock);
+			return;
+		}
+		mutex_unlock(&queue->offload_release_lock);
+	}
+
+	nvmet_rdma_free_queue(queue, !queue->offload_free_started);
+
+	kref_put(&dev->ref, nvmet_rdma_free_dev);
+}
+
 static void nvmet_rdma_release_queue_work(struct work_struct *w)
 {
 	struct nvmet_rdma_queue *queue =
 		container_of(w, struct nvmet_rdma_queue, release_work);
-	struct nvmet_rdma_device *dev = queue->dev;
 
-	nvmet_rdma_free_queue(queue);
+	nvmet_rdma_release_queue(queue);
+}
 
-	kref_put(&dev->ref, nvmet_rdma_free_dev);
+static void nvmet_rdma_release_offload_queue_work(struct work_struct *w)
+{
+	struct nvmet_rdma_queue *queue =
+		container_of(w, struct nvmet_rdma_queue, release_offload_work);
+
+	nvmet_rdma_release_queue(queue);
 }
 
 static void nvmet_rdma_disconnect_queue_work(struct work_struct *w)
@@ -1542,11 +1593,14 @@ nvmet_rdma_alloc_queue(struct nvmet_rdma_device *ndev,
 	 * inside a CM callback would trigger a deadlock. (great API design..)
 	 */
 	INIT_WORK(&queue->release_work, nvmet_rdma_release_queue_work);
+	INIT_WORK(&queue->release_offload_work,
+		  nvmet_rdma_release_offload_queue_work);
 	INIT_WORK(&queue->disconnect_work, nvmet_rdma_disconnect_queue_work);
 	queue->dev = ndev;
 	queue->cm_id = cm_id;
 	queue->port = port->nport;
 	queue->offload = queue->port->offload && queue->host_qid;
+	mutex_init(&queue->offload_release_lock);
 
 	spin_lock_init(&queue->state_lock);
 	queue->state = NVMET_RDMA_Q_CONNECTING;
@@ -1577,7 +1631,7 @@ nvmet_rdma_alloc_queue(struct nvmet_rdma_device *ndev,
 		goto out_ida_remove;
 	}
 
-	if (ndev->srqs) {
+	if (ndev->srqs && queue->host_qid) {
 		queue->nsrq = ndev->srqs[queue->comp_vector % ndev->srq_count];
 	} else if(!queue->nsrq && !queue->offload) {
 		queue->cmds = nvmet_rdma_alloc_cmds(ndev,
@@ -1635,6 +1689,11 @@ static void nvmet_rdma_qp_event(struct ib_event *event, void *priv)
 		pr_debug("received IB QP event: %s (%d)\n",
 			 ib_event_msg(event->event), event->event);
 		break;
+	case IB_EVENT_XRQ_QP_LAST_NVME_CQE_REACHED:
+		pr_debug("queue %p received IB QP event: %s (%d)\n",
+			 queue, ib_event_msg(event->event), event->event);
+		queue_work(nvmet_wq, &queue->release_offload_work);
+		break;
 	default:
 		pr_err("received IB QP event: %s (%d)\n",
 		       ib_event_msg(event->event), event->event);
@@ -1689,8 +1748,19 @@ static int nvmet_rdma_queue_connect(struct rdma_cm_id *cm_id,
 	}
 
 	if (queue->host_qid == 0) {
-		/* Let inflight controller teardown complete */
-		flush_workqueue(nvmet_wq);
+		struct nvmet_rdma_queue *q;
+		int pending = 0;
+
+		/* Check for pending controller teardown */
+		mutex_lock(&nvmet_rdma_queue_mutex);
+		list_for_each_entry(q, &nvmet_rdma_queue_list, queue_list) {
+			if (q->nvme_sq.ctrl == queue->nvme_sq.ctrl &&
+			    q->state == NVMET_RDMA_Q_DISCONNECTING)
+				pending++;
+		}
+		mutex_unlock(&nvmet_rdma_queue_mutex);
+		if (pending > NVMET_RDMA_BACKLOG)
+			return NVME_SC_CONNECT_CTRL_BUSY;
 	}
 
 	ret = nvmet_rdma_cm_accept(cm_id, queue, &event->param.conn);
@@ -1710,7 +1780,7 @@ static int nvmet_rdma_queue_connect(struct rdma_cm_id *cm_id,
 	return 0;
 
 free_queue:
-	nvmet_rdma_free_queue(queue);
+	nvmet_rdma_free_queue(queue, true);
 put_device:
 	kref_put(&ndev->ref, nvmet_rdma_free_dev);
 
@@ -1921,8 +1991,6 @@ static void nvmet_rdma_disable_port(struct nvmet_rdma_port *port)
 	if (nport->offload && cm_id)
 		nvmet_rdma_destroy_xrqs(nport);
 
- 	port->cm_id = NULL;
-
 	if (cm_id)
 		rdma_destroy_id(cm_id);
 
@@ -1963,7 +2031,7 @@ static int nvmet_rdma_enable_port(struct nvmet_rdma_port *port)
 		goto out_destroy_id;
 	}
 
-	ret = rdma_listen(cm_id, 128);
+	ret = rdma_listen(cm_id, NVMET_RDMA_BACKLOG);
 	if (ret) {
 		pr_err("listening to %pISpcs failed (%d)\n", addr, ret);
 		goto out_destroy_id;
@@ -2038,6 +2106,14 @@ static int nvmet_rdma_add_port(struct nvmet_port *nport)
 		nport->inline_data_size = NVMET_RDMA_MAX_INLINE_DATA_SIZE;
 	}
 
+	if (nport->max_queue_size < 0) {
+		nport->max_queue_size = NVME_RDMA_DEFAULT_QUEUE_SIZE;
+	} else if (nport->max_queue_size > NVME_RDMA_MAX_QUEUE_SIZE) {
+		pr_warn("max_queue_size %u is too large, reducing to %u\n",
+			nport->max_queue_size, NVME_RDMA_MAX_QUEUE_SIZE);
+		nport->max_queue_size = NVME_RDMA_MAX_QUEUE_SIZE;
+	}
+
 	ret = inet_pton_with_scope(&init_net, af, nport->disc_addr.traddr,
 			nport->disc_addr.trsvcid, &port->addr);
 	if (ret) {
@@ -2101,6 +2177,8 @@ static u8 nvmet_rdma_get_mdts(const struct nvmet_ctrl *ctrl)
 
 static u16 nvmet_rdma_get_max_queue_size(const struct nvmet_ctrl *ctrl)
 {
+	if (ctrl->pi_support)
+		return NVME_RDMA_MAX_METADATA_QUEUE_SIZE;
 	return NVME_RDMA_MAX_QUEUE_SIZE;
 }
 
@@ -2124,6 +2202,7 @@ static const struct nvmet_fabrics_ops nvmet_rdma_ops = {
 	.create_offload_ctrl	= nvmet_rdma_create_offload_ctrl,
 	.enable_offload_ns	= nvmet_rdma_enable_offload_ns,
 	.disable_offload_ns	= nvmet_rdma_disable_offload_ns,
+	.offload_ns_is_active	= nvmet_rdma_offload_ns_is_active,
 	.peer_to_peer_sqe_inline_size = nvmet_rdma_peer_to_peer_sqe_inline_size,
 	.peer_to_peer_mdts	= nvmet_rdma_peer_to_peer_mdts,
 	.offload_subsys_unknown_ns_cmds	= nvmet_rdma_offload_subsys_unknown_ns_cmds,
@@ -2205,6 +2284,23 @@ static void nvmet_rdma_remove_one(struct ib_device *ib_device, void *client_data
 
 	if (found)
 		flush_workqueue(nvmet_wq);
+
+retry:
+	/*
+	 * Wait until all devices were released. This check is needed only
+	 * for offload queues, which don't release the device before getting
+	 * last NVMe CQE events on all the queues.
+	 */
+	mutex_lock(&device_list_mutex);
+	list_for_each_entry(ndev, &device_list, entry) {
+		if (ndev->device == ib_device) {
+			mutex_unlock(&device_list_mutex);
+			msleep(100);
+			goto retry;
+		}
+	}
+	mutex_unlock(&device_list_mutex);
+	pr_info("Removed all queues of ib_device %p\n", ib_device);
 }
 
 static struct ib_client nvmet_rdma_ib_client = {
@@ -2277,6 +2373,7 @@ static void __exit nvmet_rdma_exit(void)
 module_init(nvmet_rdma_init);
 module_exit(nvmet_rdma_exit);
 
+MODULE_DESCRIPTION("NVMe target RDMA transport driver");
 MODULE_LICENSE("GPL v2");
 MODULE_ALIAS("nvmet-transport-1"); /* 1 == NVMF_TRTYPE_RDMA */
 
