@@ -223,7 +223,11 @@ static int nvme_map_user_request(struct request *req, u64 ubuffer,
 		if (WARN_ON_ONCE(flags & NVME_IOCTL_VEC))
 			return -EINVAL;
 		ret = io_uring_cmd_import_fixed(ubuffer, bufflen,
+#ifdef HAVE_IO_URING_CMD_IMPORT_FIXED_6_PARAMS
+				rq_data_dir(req), &iter, ioucmd, flags);
+#else
 				rq_data_dir(req), &iter, ioucmd);
+#endif
 		if (ret < 0)
 			goto out;
 		ret = blk_rq_map_user_iov(q, req, NULL, &iter, GFP_KERNEL);
