@@ -4,6 +4,7 @@
 #include <linux/module.h>
 #include <linux/init.h>
 #include <linux/errno.h>
+#include <linux/mlx5/driver.h>
 #include <linux/netdevice.h>
 #include <linux/rtnetlink.h>
 #include <net/genetlink.h>
@@ -38,6 +39,7 @@ static const struct nla_policy mlxdevm_function_nl_policy[MLXDEVM_PORT_FUNCTION_
 	[MLXDEVM_PORT_FN_ATTR_EXT_CAP_ROCE] = { .type = NLA_U8 },
 	[MLXDEVM_PORT_FN_ATTR_EXT_CAP_UC_LIST] = { .type = NLA_U32 },
 	[MLXDEVM_PORT_FN_ATTR_TRUST_STATE] = { .type = NLA_U8 },
+	[MLXDEVM_PORT_FN_ATTR_MAX_IO_EQS] = { .type = NLA_U32 },
 };
 
 static int mlxdevm_nl_dev_handle_fill(struct sk_buff *msg,
@@ -136,6 +138,7 @@ void mlxdevm_rate_nodes_destroy(struct mlxdevm *dev)
 {
 	const struct mlxdevm_ops *ops = dev->ops;
 	struct mlxdevm_rate_group *cur, *tmp;
+	u32 tc_bw[MLX5_MAX_NUM_TC] = {};
 	struct mlxdevm_port *port;
 
 	list_for_each_entry(port, &dev->port_list, list) {
@@ -145,6 +148,7 @@ void mlxdevm_rate_nodes_destroy(struct mlxdevm *dev)
 	}
 
 	list_for_each_entry_safe(cur, tmp, &dev->rate_group_list, list) {
+		ops->rate_node_tc_bw_set(dev, cur->name, tc_bw, NULL);
 		ops->rate_node_del(dev, cur->name, NULL);
 	}
 }
@@ -898,8 +902,7 @@ int mlxdevm_param_driverinit_value_set(struct mlxdevm *devm, u32 param_id,
 }
 EXPORT_SYMBOL_GPL(mlxdevm_param_driverinit_value_set);
 
-static struct mlxdevm_port *mlxdevm_port_get_by_index(struct mlxdevm *dev,
-						      unsigned int port_index)
+struct mlxdevm_port *mlxdevm_port_get_by_index(struct mlxdevm *dev, unsigned int port_index)
 {
 	struct mlxdevm_port *port;
 
@@ -909,6 +912,7 @@ static struct mlxdevm_port *mlxdevm_port_get_by_index(struct mlxdevm *dev,
 	}
 	return NULL;
 }
+EXPORT_SYMBOL_GPL(mlxdevm_port_get_by_index);
 
 static bool mlxdevm_port_index_exists(struct mlxdevm *dev,
 				      unsigned int port_index)
@@ -1140,6 +1144,23 @@ mlxdevm_nl_port_attrs_put(struct sk_buff *msg, struct mlxdevm_port *port)
 				attrs->pci_sf.sf))
 			return -EMSGSIZE;
 		break;
+	case MLXDEVM_PORT_FLAVOUR_PCI_PF:
+		if (nla_put_u32(msg, MLXDEVM_ATTR_PORT_CONTROLLER_NUMBER,
+				attrs->pci_pf.controller) ||
+		    nla_put_u16(msg, MLXDEVM_ATTR_PORT_PCI_PF_NUMBER, attrs->pci_pf.pf))
+			return -EMSGSIZE;
+		if (nla_put_u8(msg, MLXDEVM_ATTR_PORT_EXTERNAL, attrs->pci_pf.external))
+			return -EMSGSIZE;
+		break;
+	case MLXDEVM_PORT_FLAVOUR_PCI_VF:
+		if (nla_put_u32(msg, MLXDEVM_ATTR_PORT_CONTROLLER_NUMBER,
+				attrs->pci_vf.controller) ||
+		    nla_put_u16(msg, MLXDEVM_ATTR_PORT_PCI_PF_NUMBER, attrs->pci_vf.pf) ||
+		    nla_put_u16(msg, MLXDEVM_ATTR_PORT_PCI_VF_NUMBER, attrs->pci_vf.vf))
+			return -EMSGSIZE;
+		if (nla_put_u8(msg, MLXDEVM_ATTR_PORT_EXTERNAL, attrs->pci_vf.external))
+			return -EMSGSIZE;
+		break;
 	default:
 		break;
 	}
@@ -1220,6 +1241,10 @@ mlxdevm_port_fn_state_fill(const struct mlxdevm_ops *ops,
 	enum mlxdevm_port_fn_state state;
 	int err;
 
+	/* state and opstate are only applicable for sf */
+	if (port->attrs.flavour != MLXDEVM_PORT_FLAVOUR_PCI_SF)
+		return 0;
+
 	if (!ops->port_fn_state_get)
 		return 0;
 
@@ -1265,6 +1290,8 @@ mlxdevm_port_fn_cap_fill(const struct mlxdevm_ops *ops,
 
 	if (!ops->port_fn_cap_get)
 		return 0;
+	if (port->attrs.flavour != MLXDEVM_PORT_FLAVOUR_PCI_SF)
+		return 0;
 
 	err = ops->port_fn_cap_get(port, &cap, extack);
 	if (err) {
@@ -1291,6 +1318,32 @@ mlxdevm_port_fn_cap_fill(const struct mlxdevm_ops *ops,
 	return 0;
 }
 
+static int
+mlxdevm_port_fn_max_io_eqs_fill(const struct mlxdevm_ops *ops,
+				struct mlxdevm_port *port,
+				struct sk_buff *msg,
+				struct netlink_ext_ack *extack,
+				bool *msg_updated)
+{
+	u32 max_io_eqs;
+	int err;
+
+	if (!ops->port_fn_max_io_eqs_get)
+		return 0;
+
+	err = ops->port_fn_max_io_eqs_get(port, &max_io_eqs, extack);
+	if (err) {
+		if (err == -EOPNOTSUPP)
+			return 0;
+		return err;
+	}
+	err = nla_put_u32(msg, MLXDEVM_PORT_FN_ATTR_MAX_IO_EQS, max_io_eqs);
+	if (err)
+		return err;
+	*msg_updated = true;
+	return 0;
+}
+
 static int
 mlxdevm_nl_port_fn_attrs_put(struct sk_buff *msg,
 			     struct mlxdevm_port *port,
@@ -1323,6 +1376,10 @@ mlxdevm_nl_port_fn_attrs_put(struct sk_buff *msg,
 	if (err)
 		goto out;
 	err = mlxdevm_port_fn_trust_fill(ops, port, msg, extack, &msg_updated);
+
+	if (err)
+		goto out;
+	err = mlxdevm_port_fn_max_io_eqs_fill(ops, port, msg, extack, &msg_updated);
 out:
 	if (err || !msg_updated)
 		nla_nest_cancel(msg, fn_attr);
@@ -1462,14 +1519,30 @@ out:
 }
 
 static int mlxdevm_nl_rate_fill_common(struct sk_buff *msg, enum mlxdevm_rate_type rate_type,
-				       u64 tx_max, u64 tx_share)
+				       u64 tx_max, u64 tx_share, u32 *tc_bw)
 {
+	struct nlattr *nla_tc_bw;
+	int i;
+
 	if (nla_put_u16(msg, MLXDEVM_ATTR_EXT_RATE_TYPE, rate_type))
 		return -EMSGSIZE;
 	if (nla_put_u64_64bit(msg, MLXDEVM_ATTR_EXT_RATE_TX_MAX, tx_max, MLXDEVM_ATTR_EXT_PAD))
 		return -EMSGSIZE;
 	if (nla_put_u64_64bit(msg, MLXDEVM_ATTR_EXT_RATE_TX_SHARE, tx_share, MLXDEVM_ATTR_EXT_PAD))
 		return -EMSGSIZE;
+
+	nla_tc_bw = nla_nest_start(msg, MLXDEVM_ATTR_RATE_TC_BW);
+	if (!nla_tc_bw)
+		return -EMSGSIZE;
+
+	for (i = 0; i < IEEE_8021QAZ_MAX_TCS; i++) {
+		if (nla_put_u32(msg, i + MLXDEVM_ATTR_RATE_TC_0_BW, tc_bw[i])) {
+			nla_nest_cancel(msg, nla_tc_bw);
+			return -EMSGSIZE;
+		}
+	}
+	nla_nest_end(msg, nla_tc_bw);
+
 	return 0;
 }
 
@@ -1511,7 +1584,8 @@ static int mlxdevm_nl_leaf_fill(struct sk_buff *msg,
 			goto nla_put_failure;
 	}
 
-	if (mlxdevm_nl_rate_fill_common(msg, MLXDEVM_RATE_EXT_TYPE_LEAF, tx_max, tx_share))
+	if (mlxdevm_nl_rate_fill_common(msg, MLXDEVM_RATE_EXT_TYPE_LEAF, tx_max, tx_share,
+					port->tc_bw))
 		goto nla_put_failure;
 
 	genlmsg_end(msg, hdr);
@@ -1541,7 +1615,7 @@ static int mlxdevm_nl_node_fill(struct sk_buff *msg, struct mlxdevm *dev,
 		goto nla_put_failure;
 
 	if (mlxdevm_nl_rate_fill_common(msg, MLXDEVM_RATE_EXT_TYPE_NODE,
-					group->tx_max, group->tx_share))
+					group->tx_max, group->tx_share, group->tc_bw))
 		goto nla_put_failure;
 
 	genlmsg_end(msg, hdr);
@@ -1578,6 +1652,20 @@ static int mlxdevm_port_fn_leaf_tx_max(struct mlxdevm_port *port,
 	return ops->rate_leaf_tx_max_set(port, tx_max, extack);
 }
 
+static int mlxdevm_port_fn_leaf_tc_bw(struct mlxdevm_port *port, u32 *tc_bw,
+				      struct netlink_ext_ack *extack)
+{
+	const struct mlxdevm_ops *ops = port->devm->ops;
+
+	if (!ops || !ops->rate_leaf_tc_bw_set) {
+		NL_SET_ERR_MSG_MOD(extack,
+				   "Function does not support leaf tc bw setting");
+		return -EOPNOTSUPP;
+	}
+
+	return ops->rate_leaf_tc_bw_set(port, tc_bw, extack);
+}
+
 static int mlxdevm_port_fn_leaf_group(struct mlxdevm_port *port,
 				      const char *group, struct netlink_ext_ack *extack)
 {
@@ -1617,6 +1705,20 @@ static int mlxdevm_port_fn_node_tx_max(struct mlxdevm *dev, const char *group,
 	return ops->rate_node_tx_max_set(dev, group, tx_max, extack);
 }
 
+static int mlxdevm_port_fn_node_tc_bw(struct mlxdevm *dev, const char *group_name,
+				      u32 *tc_bw, struct netlink_ext_ack *extack)
+{
+	const struct mlxdevm_ops *ops = dev->ops;
+
+	if (!ops || !ops->rate_leaf_tc_bw_set) {
+		NL_SET_ERR_MSG_MOD(extack,
+				   "Function does not support node tc bw setting");
+		return -EOPNOTSUPP;
+	}
+
+	return ops->rate_node_tc_bw_set(dev, group_name, tc_bw, extack);
+}
+
 static int mlxdevm_rate_node_get_doit_locked(struct mlxdevm *dev, struct genl_info *info,
 					     struct sk_buff *msg)
 {
@@ -1785,11 +1887,22 @@ out:
 	return msg->len;
 }
 
+const struct nla_policy mlxdevm_rate_tc_bw_nl_policy[MLXDEVM_ATTR_RATE_TC_BW_MAX + 1] = {
+	[MLXDEVM_ATTR_RATE_TC_0_BW] = { .type = NLA_U32, },
+	[MLXDEVM_ATTR_RATE_TC_1_BW] = { .type = NLA_U32, },
+	[MLXDEVM_ATTR_RATE_TC_2_BW] = { .type = NLA_U32, },
+	[MLXDEVM_ATTR_RATE_TC_3_BW] = { .type = NLA_U32, },
+	[MLXDEVM_ATTR_RATE_TC_4_BW] = { .type = NLA_U32, },
+	[MLXDEVM_ATTR_RATE_TC_5_BW] = { .type = NLA_U32, },
+	[MLXDEVM_ATTR_RATE_TC_6_BW] = { .type = NLA_U32, },
+	[MLXDEVM_ATTR_RATE_TC_7_BW] = { .type = NLA_U32, },
+};
+
 static int mlxdevm_cmd_rate_set_leaf(struct genl_info *info, struct mlxdevm *dev)
 {
 	struct netlink_ext_ack *extack = info->extack;
+	u32 tc_bw[IEEE_8021QAZ_MAX_TCS];
 	struct mlxdevm_port *port;
-	const char *group = NULL;
 	int err = 0;
 	u64 rate;
 
@@ -1814,7 +1927,36 @@ static int mlxdevm_cmd_rate_set_leaf(struct genl_info *info, struct mlxdevm *dev
 			goto out;
 	}
 
+	if (info->attrs[MLXDEVM_ATTR_RATE_TC_BW]) {
+		struct nlattr *nla_tc_bw = info->attrs[MLXDEVM_ATTR_RATE_TC_BW];
+		struct nlattr *tb[MLXDEVM_ATTR_RATE_TC_BW_MAX + 1];
+		int i;
+
+		err = nla_parse_nested(tb, MLXDEVM_ATTR_RATE_TC_BW_MAX, nla_tc_bw,
+				       mlxdevm_rate_tc_bw_nl_policy, info->extack);
+		if (err)
+			goto out;
+
+		for (i = 0; i < IEEE_8021QAZ_MAX_TCS; i++) {
+			int j = i + MLXDEVM_ATTR_RATE_TC_0_BW;
+
+			if (tb[j]) {
+				tc_bw[i] = nla_get_u32(tb[j]);
+			} else {
+				tc_bw[i] = 0;
+			}
+		}
+
+		err = mlxdevm_port_fn_leaf_tc_bw(port, tc_bw, info->extack);
+		if (err)
+			goto out;
+
+		memcpy(port->tc_bw, tc_bw, sizeof(tc_bw));
+	}
+
 	if (info->attrs[MLXDEVM_ATTR_EXT_RATE_PARENT_NODE_NAME]) {
+		const char *group = NULL;
+
 		group = nla_strdup(info->attrs[MLXDEVM_ATTR_EXT_RATE_PARENT_NODE_NAME], GFP_KERNEL);
 		if (!group) {
 			err = -ENOMEM;
@@ -1822,14 +1964,9 @@ static int mlxdevm_cmd_rate_set_leaf(struct genl_info *info, struct mlxdevm *dev
 		}
 
 		err = mlxdevm_port_fn_leaf_group(port, group, extack);
-		if (err)
-			goto leaf_group_err;
+		kfree(group);
 	}
 
-	goto out;
-
-leaf_group_err:
-	kfree(group);
 out:
 	up_write(&dev->port_list_rwsem);
 	return err;
@@ -1861,6 +1998,29 @@ static int mlxdevm_cmd_rate_set_node(struct genl_info *info, struct mlxdevm *dev
 		err = mlxdevm_port_fn_node_tx_max(dev, group, rate, extack);
 	}
 
+	if (info->attrs[MLXDEVM_ATTR_RATE_TC_BW]) {
+		struct nlattr *nla_tc_bw = info->attrs[MLXDEVM_ATTR_RATE_TC_BW];
+		struct nlattr *tb[MLXDEVM_ATTR_RATE_TC_BW_MAX + 1];
+		u32 tc_bw[IEEE_8021QAZ_MAX_TCS];
+		int i;
+
+		err = nla_parse_nested(tb, MLXDEVM_ATTR_RATE_TC_BW_MAX, nla_tc_bw,
+				       mlxdevm_rate_tc_bw_nl_policy, info->extack);
+		if (err)
+			goto out;
+
+		for (i = 0; i < IEEE_8021QAZ_MAX_TCS; i++) {
+			int j = i + MLXDEVM_ATTR_RATE_TC_0_BW;
+
+			if (tb[j])
+				tc_bw[i] = nla_get_u32(tb[j]);
+			else
+				tc_bw[i] = 0;
+		}
+
+		err = mlxdevm_port_fn_node_tc_bw(dev, group, tc_bw, info->extack);
+	}
+
 out:
 	kfree(group);
 	return err;
@@ -2069,6 +2229,23 @@ mlxdevm_port_fn_trust_set(struct mlxdevm_port *port,
 	return ops->port_fn_trust_set(port, trust, extack);
 }
 
+static int
+mlxdevm_port_fn_max_io_eqs_set(struct mlxdevm_port *port,
+			       const struct nlattr *attr,
+			       struct netlink_ext_ack *extack)
+{
+	const struct mlxdevm_ops *ops;
+	u32 max_io_eqs;
+	max_io_eqs = nla_get_u32(attr);
+	ops = port->devm->ops;
+	if (!ops->port_fn_max_io_eqs_set) {
+		NL_SET_ERR_MSG_MOD(extack,
+				   "Function does not support max_io_eqs setting");
+		return -EOPNOTSUPP;
+	}
+	return ops->port_fn_max_io_eqs_set(port, max_io_eqs, extack);
+}
+
 static int mlxdevm_port_fn_state_set(struct mlxdevm_port *port,
 				     const struct nlattr *attr,
 				     struct netlink_ext_ack *extack)
@@ -2084,6 +2261,8 @@ static int mlxdevm_port_fn_state_set(struct mlxdevm_port *port,
 				   "Function does not support state setting");
 		return -EOPNOTSUPP;
 	}
+	if (port->attrs.flavour != MLXDEVM_PORT_FLAVOUR_PCI_SF)
+		return -EOPNOTSUPP;
 	return ops->port_fn_state_set(port, state, extack);
 }
 
@@ -2121,6 +2300,13 @@ mlxdevm_port_fn_set(struct mlxdevm_port *port,
 		if (err)
 			return err;
 	}
+
+	attr = tb[MLXDEVM_PORT_FN_ATTR_MAX_IO_EQS];
+	if (attr) {
+		err = mlxdevm_port_fn_max_io_eqs_set(port, attr, extack);
+		if (err)
+			return err;
+	}
 	/* Keep this as the last function attribute set, so that when
 	 * multiple port function attributes are set along with state,
 	 * Those can be applied first before activating the state.
@@ -2290,6 +2476,8 @@ mlxdevm_port_fn_cap_set(struct mlxdevm_port *port,
 		NL_SET_ERR_MSG_MOD(extack, "Function capability does not support setting");
 		return -EOPNOTSUPP;
 	}
+	if (port->attrs.flavour != MLXDEVM_PORT_FLAVOUR_PCI_SF)
+		return -EOPNOTSUPP;
 
 	tb = kcalloc(MLXDEVM_PORT_FUNCTION_ATTR_MAX + 1, sizeof(struct nlattr *), GFP_KERNEL);
 	if (!tb)
@@ -2462,6 +2650,7 @@ static struct genl_family mlxdevm_nl_family __ro_after_init = {
 	.netnsok = false,
 	.module = THIS_MODULE,
 	.ops = mlxdevm_nl_ops,
+	.parallel_ops = true,
 	.n_ops = ARRAY_SIZE(mlxdevm_nl_ops),
 	.resv_start_op = MLXDEVM_CMD_MAX + 1,
 };
