@@ -2,12 +2,38 @@
 #include "nnt_device_defs.h"
 #include "nnt_defs.h"
 #include "nnt_pci_conf_access.h"
-#include "nnt_pci_conf_access_no_vsec.h"
+#include "nnt_pci_conf_access_recovery.h"
 #include "nnt_memory_access.h"
+#include "nnt_gpu.h"
 #include <linux/module.h>
 
 MODULE_LICENSE("GPL");
 
+/* Forward declarations */
+int create_file_name_mstflint(struct pci_dev* pci_device, struct nnt_device* nnt_dev, enum nnt_device_type device_type);
+int create_file_name_mft(struct pci_dev* pci_device, struct nnt_device* nnt_dev, enum nnt_device_type device_type);
+int nnt_device_structure_init(struct nnt_device** nnt_device);
+int create_nnt_device(struct pci_dev* pci_device, enum nnt_device_type device_type, int is_alloc_chrdev_region);
+int check_pci_id_range(unsigned short pci_device_id, unsigned short id_range_start);
+int is_connectx(unsigned short pci_device_id);
+int is_connectx3(unsigned short pci_device_id);
+int is_bluefield(unsigned short pci_device_id);
+int is_pcie_switch(unsigned short pci_device_id);
+int is_quantum(unsigned short pci_device_id);
+int is_spectrum(unsigned short pci_device_id);
+int is_switch_ib(unsigned short pci_device_id);
+int is_livefish_device(unsigned short pci_device_id);
+int is_nic(unsigned short pci_device_id);
+int is_switch(unsigned short pci_device_id);
+int is_toolspf(unsigned short pci_device_id);
+int create_device_file(struct nnt_device* current_nnt_device,
+                       dev_t device_number,
+                       int minor,
+                       struct file_operations* fop,
+                       int is_alloc_chrdev_region);
+int create_devices(dev_t device_number, struct file_operations* fop, int is_alloc_chrdev_region);
+int find_all_vendor_devices(unsigned int vendor_id);
+
 /* Device list to check if device is available
      since it could be removed by hotplug event. */
 LIST_HEAD(nnt_device_list);
@@ -23,6 +49,12 @@ int get_nnt_device(struct file* file, struct nnt_device** nnt_device)
     else
     {
         *nnt_device = file->private_data;
+
+        if (!(*nnt_device)->pci_device)
+        {
+            printk(KERN_ERR "mst_pciconf driver: pci_device pointer is NULL\n");
+            error_code = -EINVAL;
+        }
     }
 
     return error_code;
@@ -192,7 +224,7 @@ ReturnOnError:
 
 int check_pci_id_range(unsigned short pci_device_id, unsigned short id_range_start)
 {
-    return (pci_device_id >= id_range_start) && (pci_device_id <= (id_range_start + 100));
+    return (pci_device_id >= id_range_start) && (pci_device_id <= (id_range_start + 255));
 }
 
 int is_connectx(unsigned short pci_device_id)
@@ -231,16 +263,6 @@ int is_switch_ib(unsigned short pci_device_id)
     return pci_device_id == SWITCHIB_PCI_ID || pci_device_id == SWITCHIB2_PCI_ID;
 }
 
-int is_bw00(unsigned short pci_device_id)
-{
-    return check_pci_id_range(pci_device_id, BW00_PCI_ID);
-}
-
-int is_bw02(unsigned short pci_device_id)
-{
-    return check_pci_id_range(pci_device_id, BW02_PCI_ID);
-}
-
 int is_livefish_device(unsigned short pci_device_id)
 {
     return pci_device_id >= CONNECTX3_LIVEFISH_ID && pci_device_id < CONNECTX3_PCI_ID;
@@ -254,7 +276,7 @@ int is_nic(unsigned short pci_device_id)
 int is_switch(unsigned short pci_device_id)
 {
     return is_pcie_switch(pci_device_id) || is_quantum(pci_device_id) || is_spectrum(pci_device_id) ||
-           is_switch_ib(pci_device_id) || is_bw00(pci_device_id) || is_bw02(pci_device_id);
+           is_switch_ib(pci_device_id);
 }
 
 int is_toolspf(unsigned short pci_device_id)
@@ -264,14 +286,17 @@ int is_toolspf(unsigned short pci_device_id)
 
 int is_pciconf_device(unsigned short pci_device_id)
 {
-    return is_nic(pci_device_id) || is_toolspf(pci_device_id) || is_livefish_device(pci_device_id) ||
-           is_switch(pci_device_id);
+    int result = is_nic(pci_device_id) || is_toolspf(pci_device_id) || is_livefish_device(pci_device_id) ||
+                 is_switch(pci_device_id) || is_gpu_pci_device(pci_device_id);
+    return result;
 }
 
 int is_pcicr_device(unsigned short pci_device_id)
 {
-    return (is_switch(pci_device_id) || is_toolspf(pci_device_id) || is_connectx3(pci_device_id)) &&
-           (!is_livefish_device(pci_device_id));
+    int result = (is_gpu_pci_device(pci_device_id) || is_switch(pci_device_id) || is_toolspf(pci_device_id) ||
+                  is_connectx3(pci_device_id)) &&
+                 (!is_livefish_device(pci_device_id));
+    return result;
 }
 
 int create_device_file(struct nnt_device* current_nnt_device,
@@ -328,67 +353,56 @@ ReturnOnFinished:
     return error;
 }
 
-int check_if_vsec_supported(struct nnt_device* nnt_device)
+unsigned int get_vsc_address_by_type(struct nnt_device* current_nnt_device, int vsc_type)
 {
-    int error = 0;
+    unsigned int vsc_address = 0;
 
-    error = nnt_device->access.init(nnt_device);
-    CHECK_ERROR(error);
+    // Attempt to find the first Vendor-Specific Capability (VSC)
+    vsc_address = pci_find_capability(current_nnt_device->pci_device, VENDOR_SPECIFIC_CAPABILITY_ID);
 
-    if (!nnt_device->pciconf_device.vsec_fully_supported)
+    // Check if the first VSC is of the specified type
+    if (vsc_address && is_vsc_type(current_nnt_device, vsc_address, vsc_type))
     {
-        nnt_device->device_type = NNT_PCICONF_NO_VSEC;
-        nnt_device->access.read = read_pciconf_no_vsec;
-        nnt_device->access.write = write_pciconf_no_vsec;
-        nnt_device->access.init = init_pciconf_no_vsec;
+        // printk(KERN_DEBUG "Device ID :%d has VSC of type: %d at offset:%x of the config space\n",
+        //        current_nnt_device->pci_device->device, vsc_type, vsc_address);
+        return vsc_address;
     }
 
-ReturnOnFinished:
-    return error;
+    // Iterate through the capabilities linked list to find the specified VSC type
+    while ((vsc_address =
+              pci_find_next_capability(current_nnt_device->pci_device, vsc_address, VENDOR_SPECIFIC_CAPABILITY_ID)))
+    {
+        if (is_vsc_type(current_nnt_device, vsc_address, vsc_type))
+        {
+            // printk(KERN_DEBUG "Device ID :%d has VSC of type: %d at offset:%x of the config space\n",
+            //        current_nnt_device->pci_device->device, vsc_type, vsc_address);
+            return vsc_address;
+        }
+    }
+
+    // printk(KERN_DEBUG "No VSC of type: %d found for device ID: %d\n", vsc_type,
+    // current_nnt_device->pci_device->device);
+    return 0; // Return 0 if no VSC of the specified type is found
 }
 
-int is_mellanox_vendor_type(struct nnt_device* current_nnt_device, unsigned int vsec_address)
+/* Return 1 if VSC is of type vsc_type */
+int is_vsc_type(struct nnt_device* current_nnt_device, unsigned int vsec_address, unsigned int vsc_type)
 {
-    u_int8_t vendor_type = 0;
+    u_int8_t type = 0;
     int error = 0;
     /* Read the capability type field */
-    if ((error = pci_read_config_byte(current_nnt_device->pci_device, (vsec_address + PCI_TYPE_OFFSET), &vendor_type)))
+    if ((error = pci_read_config_byte(current_nnt_device->pci_device, (vsec_address + PCI_TYPE_OFFSET), &type)))
     {
-        printk(KERN_ERR "Reading VSEC type failed with error %d\n", error);
-        return 0;
+        printk(KERN_ERR "Reading VSC type failed with error %d\n", error);
+        return -1;
     }
-
-    if (vendor_type == MELLANOX_VSEC_TYPE)
+    if (type == vsc_type)
     {
         return 1;
     }
     return 0;
 }
 
-unsigned int get_mellanox_vsec_address(struct nnt_device* current_nnt_device)
-{
-    unsigned int vsec_address = 0;
-
-    /* Look for the Mellanox VSEC address. if Mellanox VSEC isn't supported the address should be 0 */
-    vsec_address = pci_find_capability(current_nnt_device->pci_device, VSEC_CAPABILITY_ADDRESS);
-    if (!vsec_address || is_mellanox_vendor_type(current_nnt_device, vsec_address))
-    {
-        return vsec_address;
-    }
-
-    /* if found a non-Mellanox type VSEC, iterate of the next available VSECs*/
-    while (
-      (vsec_address = pci_find_next_capability(current_nnt_device->pci_device, vsec_address, VSEC_CAPABILITY_ADDRESS)))
-    {
-        if (is_mellanox_vendor_type(current_nnt_device, vsec_address))
-        {
-            return vsec_address;
-        }
-    }
-    /* if Mellanox VSEC address was not found, return 0 */
-    return 0;
-}
-
 int create_devices(dev_t device_number, struct file_operations* fop, int is_alloc_chrdev_region)
 {
     struct nnt_device* current_nnt_device = NULL;
@@ -404,13 +418,32 @@ int create_devices(dev_t device_number, struct file_operations* fop, int is_allo
             goto ReturnOnFinished;
 
         /* Members initialization. */
-        current_nnt_device->pciconf_device.vendor_specific_capability = get_mellanox_vsec_address(current_nnt_device);
+        current_nnt_device->pciconf_device.functional_vsc_offset = get_vsc_address_by_type(
+          current_nnt_device, FUNCTIONAL_VSC); // Exists when device is in Functional/Zombiefish mode
+        if (!current_nnt_device->pciconf_device.functional_vsc_offset)
+        {
+            current_nnt_device->pciconf_device.recovery_vsc_offset =
+              get_vsc_address_by_type(current_nnt_device, RECOVERY_VSC);
+            if (current_nnt_device->pciconf_device.recovery_vsc_offset != 0 &&
+                current_nnt_device->pciconf_device.recovery_vsc_offset != RECOVERY_VSC_OFFSET_IN_CONFIG_SPACE)
+            {
+                printk(
+                  KERN_ERR
+                  "Found that recovery VSC offset is: %d. Which does not allign with definition which is 0x54 or 0.\n",
+                  current_nnt_device->pciconf_device.recovery_vsc_offset);
+                return -1;
+            }
+        }
         current_nnt_device->vpd_capability_address =
           pci_find_capability(current_nnt_device->pci_device, PCI_CAP_ID_VPD);
 
-        if (!current_nnt_device->pciconf_device.vendor_specific_capability)
+        /* NNT_PCICONF_RECOVERY is for LF and late LF. */
+        if (!current_nnt_device->pciconf_device.functional_vsc_offset)
         {
-            current_nnt_device->device_type = NNT_PCICONF_NO_VSEC;
+            current_nnt_device->device_type = NNT_PCICONF_RECOVERY;
+            // printk(KERN_DEBUG
+            //        "Device with device ID: %d is in LF or Late LF mode. current_nnt_device->device_type set to:
+            //        %d\n", current_nnt_device->pci_device->device, current_nnt_device->device_type);
         }
 
         switch (current_nnt_device->device_type)
@@ -419,12 +452,9 @@ int create_devices(dev_t device_number, struct file_operations* fop, int is_allo
                 current_nnt_device->access.read = read_pciconf;
                 current_nnt_device->access.write = write_pciconf;
                 current_nnt_device->access.init = init_pciconf;
-
-                error = check_if_vsec_supported(current_nnt_device);
-                CHECK_ERROR(error);
                 break;
 
-            case NNT_PCICONF_NO_VSEC:
+            case NNT_PCICONF_RECOVERY:
                 current_nnt_device->access.read = read_pciconf_no_vsec;
                 current_nnt_device->access.write = write_pciconf_no_vsec;
                 current_nnt_device->access.init = init_pciconf_no_vsec;
@@ -624,3 +654,51 @@ int destroy_nnt_device_bc(struct nnt_device* nnt_device)
 
     return 0;
 }
+
+int rescan(void)
+{
+    struct nnt_device* current_nnt_device = NULL;
+    struct nnt_device* temp_nnt_device = NULL;
+    struct pci_dev* new_pci_device;
+
+    list_for_each_entry_safe(current_nnt_device, temp_nnt_device, &nnt_device_list, entry)
+    {
+        /* Check if rescan is needed */
+        if (current_nnt_device->pci_device->error_state == pci_channel_io_normal)
+        {
+            printk(KERN_DEBUG "No need to rescan for device %s as its error state is in pci_channel_io_normal\n",
+                   current_nnt_device->device_name);
+            continue;
+        }
+
+        new_pci_device = NULL;
+        while ((new_pci_device = pci_get_device(NNT_MELLANOX_PCI_VENDOR, current_nnt_device->pci_device->device,
+                                                new_pci_device)) != NULL)
+        {
+            /* Checking if this is the device that we need to replace. */
+            if (new_pci_device->device == current_nnt_device->pci_device->device &&
+                PCI_FUNC(new_pci_device->devfn) == PCI_FUNC(current_nnt_device->pci_device->devfn) &&
+                new_pci_device->bus->number == current_nnt_device->pci_device->bus->number &&
+                pci_domain_nr(new_pci_device->bus) == pci_domain_nr(current_nnt_device->pci_device->bus))
+            {
+                printk(
+                  KERN_DEBUG
+                  "A new instance of the pci_dev structure has been discovered. new  pointer address: %p, old pointer address: %p, mst device name: %s\n",
+                  new_pci_device, current_nnt_device->pci_device, current_nnt_device->device_name);
+
+                if (new_pci_device->error_state != pci_channel_io_normal)
+                {
+                    printk(
+                      KERN_DEBUG
+                      "The new instance of the pci_dev structure is also not in pci_channel_io_normal, error_state = %d.\n",
+                      new_pci_device->error_state);
+                }
+
+                current_nnt_device->pci_device = new_pci_device;
+                break;
+            }
+        }
+    }
+
+    return 0;
+}
