@@ -81,7 +81,13 @@ int read_dword(struct nnt_read_dword_from_config_space* read_from_cspace, struct
 
     /* Take semaphore. */
     error = lock_vsec_semaphore(nnt_device);
-    CHECK_ERROR(error);
+    if (error)
+    {
+        printk("Failed to lock VSEC semaphore for device: %04x:%02x:%02x.%d. \n",
+               pci_domain_nr(nnt_device->pci_device->bus), nnt_device->pci_device->bus->number,
+               PCI_SLOT(nnt_device->pci_device->devfn), PCI_FUNC(nnt_device->pci_device->devfn));
+        return error;
+    }
 
     /* Read dword from config space. */
     error = pci_read_config_dword(nnt_device->pci_device, read_from_cspace->offset, &read_from_cspace->data);
@@ -206,7 +212,13 @@ int check_address_space_support(struct nnt_device* nnt_device)
 
     /* Get semaphore ticket */
     error = lock_vsec_semaphore(nnt_device);
-    CHECK_ERROR(error);
+    if (error)
+    {
+        printk("Failed to lock VSEC semaphore for device: %04x:%02x:%02x.%d. \n",
+               pci_domain_nr(nnt_device->pci_device->bus), nnt_device->pci_device->bus->number,
+               PCI_SLOT(nnt_device->pci_device->devfn), PCI_FUNC(nnt_device->pci_device->devfn));
+        return error;
+    }
 
     /* Is ICMD address space supported ?*/
     if (set_address_space(nnt_device, NNT_SPACE_ALL_ICMD) == 0)
@@ -249,7 +261,7 @@ int check_address_space_support(struct nnt_device* nnt_device)
     {
         nnt_device->pciconf_device.address_space.pci_semaphore = 1;
     }
-ReturnOnFinished:
+
     /* Clear semaphore. */
     clear_vsec_semaphore(nnt_device);
 
@@ -305,7 +317,13 @@ int read_pciconf(struct nnt_device* nnt_device, struct nnt_rw_operation* read_op
 
     /* Lock semaphore. */
     error = lock_vsec_semaphore(nnt_device);
-    CHECK_ERROR(error);
+    if (error)
+    {
+        printk("Failed to lock VSEC semaphore for device: %04x:%02x:%02x.%d. \n",
+               pci_domain_nr(nnt_device->pci_device->bus), nnt_device->pci_device->bus->number,
+               PCI_SLOT(nnt_device->pci_device->devfn), PCI_FUNC(nnt_device->pci_device->devfn));
+        return error;
+    }
 
     /* Is CR Space address space supported ?*/
     error = set_address_space(nnt_device, read_operation->address_space);
@@ -360,7 +378,13 @@ int write_pciconf(struct nnt_device* nnt_device, struct nnt_rw_operation* write_
 
     /* Lock semaphore. */
     error = lock_vsec_semaphore(nnt_device);
-    CHECK_ERROR(error);
+    if (error)
+    {
+        printk("Failed to lock VSEC semaphore for device: %04x:%02x:%02x.%d. \n",
+               pci_domain_nr(nnt_device->pci_device->bus), nnt_device->pci_device->bus->number,
+               PCI_SLOT(nnt_device->pci_device->devfn), PCI_FUNC(nnt_device->pci_device->devfn));
+        return error;
+    }
 
     /* Is CR Space address space supported ?*/
     error = set_address_space(nnt_device, write_operation->address_space);
@@ -441,7 +465,13 @@ int init_vsec_capability_mask(struct nnt_device* nnt_device)
 
     /* Lock semaphore. */
     error = lock_vsec_semaphore(nnt_device);
-    CHECK_ERROR(error);
+    if (error)
+    {
+        printk("Failed to lock VSEC semaphore for device: %04x:%02x:%02x.%d. \n",
+               pci_domain_nr(nnt_device->pci_device->bus), nnt_device->pci_device->bus->number,
+               PCI_SLOT(nnt_device->pci_device->devfn), PCI_FUNC(nnt_device->pci_device->devfn));
+        return error;
+    }
 
     get_space_support_status(nnt_device, NNT_SPACE_ICMD);
     get_space_support_status(nnt_device, NNT_SPACE_CR_SPACE);
@@ -460,7 +490,6 @@ int init_vsec_capability_mask(struct nnt_device* nnt_device)
     get_space_support_status(nnt_device, NNT_SPACE_PCI_GLOBAL_SEMAPHORE);
     nnt_device->pciconf_device.vsec_capability_mask |= (1 << NNT_VSEC_INITIALIZED);
 
-ReturnOnFinished:
     /* Clear semaphore. */
     clear_vsec_semaphore(nnt_device);
 
