@@ -1,5 +1,7 @@
 /*
- * Copyright (c) 2001-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
+ * Copyright (c) 2001-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-License-Identifier: GPL-2.0-only or BSD-3-Clause
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -47,22 +49,22 @@
 // When Striding RQ is on, it points to g_buffer_pool_rx_stride since the upper layers work with
 // strides. When Striding RQ is off, it points to g_buffer_pool_rx_rwqe since the upper layers work
 // with RWQEs buffers themselves.
-buffer_pool *g_buffer_pool_rx_ptr = NULL;
+buffer_pool *g_buffer_pool_rx_ptr = nullptr;
 
 // This buffer-pool holds buffer descriptors which represent strides in strided RWQEs.
 // These buffers descriptos do not actually own a buffer.
 // Each such descriptor points into a portion of a buffer of a g_buffer_pool_rx_rwqe descriptor.
-buffer_pool *g_buffer_pool_rx_stride = NULL;
+buffer_pool *g_buffer_pool_rx_stride = nullptr;
 
 // This buffer-pool holds the actual buffers for receive WQEs.
-buffer_pool *g_buffer_pool_rx_rwqe = NULL;
+buffer_pool *g_buffer_pool_rx_rwqe = nullptr;
 
 // This buffer-pool holds the actual buffers for send WQEs.
-buffer_pool *g_buffer_pool_tx = NULL;
+buffer_pool *g_buffer_pool_tx = nullptr;
 
 // This buffer-pool holds buffer descriptors for zero copy TX.
 // These buffer descriptors do not actually own a buffer.
-buffer_pool *g_buffer_pool_zc = NULL;
+buffer_pool *g_buffer_pool_zc = nullptr;
 
 // inlining a function only help in case it come before using it...
 inline void buffer_pool::put_buffer_helper(mem_buf_desc_t *buff)
@@ -74,15 +76,15 @@ inline void buffer_pool::put_buffer_helper(mem_buf_desc_t *buff)
     }
 #endif
 
-    if (buff->lwip_pbuf.pbuf.desc.attr == PBUF_DESC_STRIDE) {
-        mem_buf_desc_t *rwqe = reinterpret_cast<mem_buf_desc_t *>(buff->lwip_pbuf.pbuf.desc.mdesc);
+    if (buff->lwip_pbuf.desc.attr == PBUF_DESC_STRIDE) {
+        mem_buf_desc_t *rwqe = reinterpret_cast<mem_buf_desc_t *>(buff->lwip_pbuf.desc.mdesc);
         if (buff->rx.strides_num == rwqe->add_ref_count(-buff->rx.strides_num)) { // Is last stride.
             g_buffer_pool_rx_rwqe->put_buffers_thread_safe(rwqe);
         }
     }
 
     buff->p_next_desc = m_p_head;
-    assert(buff->lwip_pbuf.pbuf.type != PBUF_ZEROCOPY || this == g_buffer_pool_zc ||
+    assert(buff->lwip_pbuf.type != PBUF_ZEROCOPY || this == g_buffer_pool_zc ||
            g_buffer_pool_zc == NULL);
     free_lwip_pbuf(&buff->lwip_pbuf);
     m_p_head = buff;
@@ -120,10 +122,8 @@ bool buffer_pool::expand(size_t count)
     }
 
     for (size_t i = 0; i < count; ++i) {
-        pbuf_type type = (m_buf_size == 0 && m_custom_free_function == free_tx_lwip_pbuf_custom)
-            ? PBUF_ZEROCOPY
-            : PBUF_RAM;
-        desc = new (desc_ptr) mem_buf_desc_t(data_ptr, m_buf_size, type, m_custom_free_function);
+        pbuf_type type = (m_buf_size == 0 && m_p_bpool_stat->is_tx) ? PBUF_ZEROCOPY : PBUF_RAM;
+        desc = new (desc_ptr) mem_buf_desc_t(data_ptr, m_buf_size, type);
         put_buffer_helper(desc);
         desc_ptr += sizeof(mem_buf_desc_t);
         if (data_ptr) {
@@ -131,6 +131,7 @@ bool buffer_pool::expand(size_t count)
         }
     }
     m_n_buffers_created += count;
+    m_p_bpool_stat->n_buffer_pool_created = m_n_buffers_created;
     return true;
 }
 
@@ -149,8 +150,7 @@ void buffer_pool::free_tx_lwip_pbuf_custom(struct pbuf *p_buff)
     pool->put_buffers_thread_safe((mem_buf_desc_t *)p_buff);
 }
 
-buffer_pool::buffer_pool(buffer_pool_type type, size_t buf_size,
-                         pbuf_free_custom_fn custom_free_function, alloc_t alloc_func,
+buffer_pool::buffer_pool(buffer_pool_type type, size_t buf_size, alloc_t alloc_func,
                          free_t free_func)
     : m_lock("buffer_pool")
     , m_buf_size((buf_size + MCE_ALIGNMENT) & (~MCE_ALIGNMENT))
@@ -161,7 +161,6 @@ buffer_pool::buffer_pool(buffer_pool_type type, size_t buf_size,
     , m_allocator_data(m_buf_size ? xlio_allocator_heap(alloc_func, free_func, true)
                                   : xlio_allocator_heap(false))
     , m_allocator_metadata(false)
-    , m_custom_free_function(custom_free_function)
 {
     size_t initial_pool_size;
 
@@ -220,37 +219,30 @@ void buffer_pool::print_report(vlog_levels_t log_level /*=VLOG_DEBUG*/)
     vlog_printf(log_level, "Buffer pool %p (%s%s):\n", this, m_p_bpool_stat->is_rx ? "Rx" : "Tx",
                 m_buf_size ? "" : ", zcopy");
     vlog_printf(log_level, "  Buffers: %zu created, %zu free\n", m_n_buffers_created, m_n_buffers);
-    vlog_printf(log_level, "  Memory consumption: %s (%s per buffer)\n",
+    vlog_printf(log_level, "  Memory consumption: %s (%s per buffer), expanded %u times\n",
                 option_size::to_str(m_buf_size * m_n_buffers_created, str1, sizeof(str1)),
-                option_size::to_str(m_buf_size, str2, sizeof(str2)));
+                option_size::to_str(m_buf_size, str2, sizeof(str2)),
+                m_p_bpool_stat->n_buffer_pool_expands);
     vlog_printf(log_level, "  Requests: %u unsatisfied buffer requests\n",
                 m_p_bpool_stat->n_buffer_pool_no_bufs);
 }
 
 /* static */
-void buffer_pool::print_report_on_errors(vlog_levels_t log_level)
+void buffer_pool::print_full_report(vlog_levels_t log_level)
 {
     std::vector<buffer_pool *> pools = {g_buffer_pool_rx_rwqe, g_buffer_pool_rx_stride,
                                         g_buffer_pool_tx, g_buffer_pool_zc};
-    bool do_print = false;
+    bool is_error = false;
 
     for (auto &pool : pools) {
-        if (pool->m_p_bpool_stat->n_buffer_pool_no_bufs) {
-            do_print = true;
-            break;
-        }
+        is_error = is_error || pool->m_p_bpool_stat->n_buffer_pool_no_bufs;
+        pool->print_report(log_level);
     }
 
-    if (do_print) {
+    if (is_error) {
         vlog_printf(log_level,
                     "XLIO detected insufficient memory. Increasing XLIO_MEMORY_LIMIT can improve "
                     "performance.\n");
-        for (auto &pool : pools) {
-            // Print report for every non-zcopy buffer pool and zcopy pools with errors.
-            if (pool->m_buf_size || pool->m_p_bpool_stat->n_buffer_pool_no_bufs) {
-                pool->print_report(log_level);
-            }
-        }
     }
 }
 
@@ -285,7 +277,7 @@ bool buffer_pool::get_buffers_thread_safe(descq_t &pDeque, ring_slave *desc_owne
         // Remove from list
         head = m_p_head;
         m_p_head = m_p_head->p_next_desc;
-        head->p_next_desc = NULL;
+        head->p_next_desc = nullptr;
 
         // Init
         head->lkey = lkey;
@@ -494,7 +486,7 @@ void buffer_pool::put_buffers_after_deref_thread_safe(descq_t *pDeque)
     std::lock_guard<decltype(m_lock)> lock(m_lock);
     while (!pDeque->empty()) {
         mem_buf_desc_t *list = pDeque->get_and_pop_front();
-        if (list->dec_ref_count() <= 1 && (list->lwip_pbuf.pbuf.ref-- <= 1)) {
+        if (likely(list) && list->dec_ref_count() <= 1 && (list->lwip_pbuf.ref-- <= 1)) {
             put_buffers(list);
         }
     }
@@ -502,7 +494,7 @@ void buffer_pool::put_buffers_after_deref_thread_safe(descq_t *pDeque)
 
 void buffer_pool::put_buffer_after_deref_thread_safe(mem_buf_desc_t *buff)
 {
-    if (buff->dec_ref_count() <= 1 && (buff->lwip_pbuf.pbuf.ref-- <= 1)) {
+    if (buff->dec_ref_count() <= 1 && (buff->lwip_pbuf.ref-- <= 1)) {
         std::lock_guard<decltype(m_lock)> lock(m_lock);
         put_buffers(buff);
     }
