@@ -180,29 +180,6 @@ void L3_level_tcp_input(struct pbuf *p, struct tcp_pcb *pcb)
             in_data.recv_data = NULL;
             in_data.recv_flags = 0;
 
-            /* If there is data which was previously "refused" by upper layer */
-            /* 'while' instead of 'if' because windows scale uses large pbuf */
-            while (pcb->refused_data != NULL) {
-                struct pbuf *rest;
-                pbuf_split_64k(pcb->refused_data, &rest);
-
-                /* Notify again application with data previously received. */
-                LWIP_DEBUGF(TCP_INPUT_DEBUG, ("tcp_input: notify kept packet\n"));
-                TCP_EVENT_RECV(pcb, pcb->refused_data, ERR_OK, err);
-                if (err == ERR_OK) {
-                    pcb->refused_data = rest;
-                } else {
-                    if (rest) {
-                        pbuf_cat(pcb->refused_data, rest); /* undo splitting */
-                    }
-                    /* if err == ERR_ABRT, 'pcb' is already deallocated */
-                    /* drop incoming packets, because pcb is "full" */
-                    LWIP_DEBUGF(TCP_INPUT_DEBUG,
-                                ("tcp_input: drop incoming packets, because pcb is \"full\"\n"));
-                    pbuf_free(p);
-                    return;
-                }
-            }
             pcb->is_in_input = 1;
             err = tcp_process(pcb, &in_data);
             /* A return value of ERR_ABRT means that tcp_abort() was called
@@ -212,7 +189,8 @@ void L3_level_tcp_input(struct pbuf *p, struct tcp_pcb *pcb)
                     /* TF_RESET means that the connection was reset by the other
                        end. We then call the error callback to inform the
                        application that the connection is dead before we
-                       deallocate the PCB. */
+                       deallocate the PCB.
+                       Error callback will trigger event only if pcb is active. */
                     TCP_EVENT_ERR(pcb->errf, pcb->my_container, ERR_RST);
                     tcp_pcb_remove(pcb);
                 } else if (in_data.recv_flags & TF_CLOSED) {
@@ -230,9 +208,7 @@ void L3_level_tcp_input(struct pbuf *p, struct tcp_pcb *pcb)
                         }
                     }
 
-                    while (in_data.recv_data !=
-                           NULL) { // 'while' instead of 'if' because windows scale uses large pbuf
-                        struct pbuf *rest = NULL;
+                    if (in_data.recv_data) {
                         if (pcb->flags & TF_RXCLOSED) {
                             /* received data although already closed -> abort (send RST) to
                                            notify the remote host that not all data has been
@@ -241,30 +217,18 @@ void L3_level_tcp_input(struct pbuf *p, struct tcp_pcb *pcb)
                             tcp_abort(pcb);
                             goto aborted;
                         }
-                        pbuf_split_64k(in_data.recv_data, &rest);
                         if (in_data.flags & TCP_PSH) {
                             in_data.recv_data->flags |= PBUF_FLAG_PUSH;
                         }
                         /* Notify application that data has been received. */
                         TCP_EVENT_RECV(pcb, in_data.recv_data, ERR_OK, err);
                         if (err == ERR_ABRT) {
-                            if (rest) {
-                                pbuf_cat(in_data.recv_data, rest); /* undo splitting */
-                            }
                             goto aborted;
                         }
                         /* If the upper layer can't receive this data, store it */
                         if (err != ERR_OK) {
-                            if (rest) {
-                                pbuf_cat(in_data.recv_data, rest); /* undo splitting */
-                            }
-                            pcb->refused_data = in_data.recv_data;
-                            LWIP_DEBUGF(
-                                TCP_INPUT_DEBUG,
-                                ("tcp_input: keep incoming packet, because pcb is \"full\"\n"));
-                            break;
-                        } else {
-                            in_data.recv_data = rest;
+                            pcb->rcv_wnd += in_data.recv_data->tot_len;
+                            pbuf_free(in_data.recv_data);
                         }
                     }
 
@@ -285,11 +249,6 @@ void L3_level_tcp_input(struct pbuf *p, struct tcp_pcb *pcb)
                     pcb->is_in_input = 0;
                     /* Try to send something out. */
                     tcp_output(pcb);
-#if TCP_INPUT_DEBUG
-#if TCP_DEBUG
-                    tcp_debug_print_state(get_tcp_state(pcb));
-#endif /* TCP_DEBUG */
-#endif /* TCP_INPUT_DEBUG */
                 }
             }
             /* Jump target if pcb has been aborted in a callback (by calling tcp_abort()).
@@ -298,7 +257,7 @@ void L3_level_tcp_input(struct pbuf *p, struct tcp_pcb *pcb)
             pcb->is_in_input = 0;
             in_data.recv_data = NULL;
 
-            /* give up our reference to inseg.p */
+            /* tcp_receive() sets in_data.inseg.p to NULL in case of recv_data */
             if (in_data.inseg.p != NULL) {
                 pbuf_free(in_data.inseg.p);
                 in_data.inseg.p = NULL;
@@ -712,6 +671,7 @@ static err_t tcp_process(struct tcp_pcb *pcb, tcp_in_data *in_data)
             /* Looks like another copy of the SYN - retransmit our SYN-ACK */
             tcp_rexmit(pcb);
         }
+        // Currently received packets with only FIN bit in SYN_RCVD state are ignored.
         break;
     case CLOSE_WAIT:
         /* FALLTHROUGH */
@@ -833,13 +793,10 @@ static u32_t tcp_shrink_segment(struct tcp_pcb *pcb, struct tcp_seg *seg, u32_t
     u8_t optflags = 0;
     u8_t optlen;
 
+    assert(seg != NULL);
+    assert(seg->p != NULL);
     assert(!(seg->flags & TF_SEG_OPTS_ZEROCOPY));
 
-    if ((NULL == seg) || (NULL == seg->p) ||
-        !(TCP_SEQ_GT(ackno, seg->seqno) && TCP_SEQ_LT(ackno, seg->seqno + TCP_SEGLEN(seg)))) {
-        return count;
-    }
-
 #if LWIP_TCP_TIMESTAMPS
     if ((pcb->flags & TF_TIMESTAMP)) {
         optflags |= TF_SEG_OPTS_TS;
@@ -962,10 +919,6 @@ static u32_t tcp_shrink_zc_segment(struct tcp_pcb *pcb, struct tcp_seg *seg, u32
     assert(seg->p != NULL);
     assert(seg->flags & TF_SEG_OPTS_ZEROCOPY);
 
-    if (!(TCP_SEQ_GT(ackno, seg->seqno) && TCP_SEQ_LT(ackno, seg->seqno + TCP_SEGLEN(seg)))) {
-        return 0;
-    }
-
     while (TCP_SEQ_GEQ(ackno, seg->seqno + seg->p->len)) {
         p = seg->p;
         seg->len -= p->len;
@@ -989,6 +942,32 @@ static u32_t tcp_shrink_zc_segment(struct tcp_pcb *pcb, struct tcp_seg *seg, u32
     return count;
 }
 
+static void ack_partial_or_whole_segment(struct tcp_pcb *pcb, u32_t ackno, struct tcp_seg **seg)
+{
+    struct tcp_seg *whole_seg_to_ack;
+    while ((*seg) != NULL && TCP_SEQ_GT(ackno, (*seg)->seqno)) {
+        if (TCP_SEQ_LT(ackno, (*seg)->seqno + TCP_SEGLEN((*seg)))) {
+            // Ack partial TCP segment
+            u32_t removed = (*seg)->flags & TF_SEG_OPTS_ZEROCOPY
+                ? tcp_shrink_zc_segment(pcb, (*seg), ackno)
+                : tcp_shrink_segment(pcb, (*seg), ackno);
+            pcb->snd_queuelen -= removed;
+            break;
+        }
+
+        whole_seg_to_ack = (*seg);
+        (*seg) = (*seg)->next;
+
+        /* Prevent ACK for FIN to generate a sent event */
+        if ((pcb->acked != 0) && ((whole_seg_to_ack->tcp_flags & TCP_FIN) != 0)) {
+            pcb->acked--;
+        }
+
+        pcb->snd_queuelen -= pbuf_clen(whole_seg_to_ack->p);
+        tcp_tx_seg_free(pcb, whole_seg_to_ack);
+    }
+}
+
 /**
  * Called by tcp_process. Checks if the given segment is an ACK for outstanding
  * data, and if so frees the memory of the buffered data. Next, is places the
@@ -1008,7 +987,6 @@ static void tcp_receive(struct tcp_pcb *pcb, tcp_in_data *in_data)
     struct tcp_seg *prev, *cseg;
 #endif /* TCP_QUEUE_OOSEQ */
     struct pbuf *p;
-    s32_t off;
     s16_t m;
     u32_t right_wnd_edge;
     u32_t new_tot_len;
@@ -1016,6 +994,9 @@ static void tcp_receive(struct tcp_pcb *pcb, tcp_in_data *in_data)
     s8_t persist = 0;
 
     if (in_data->flags & TCP_ACK) {
+        if (pcb->unacked) {
+            __builtin_prefetch(pcb->unacked->p);
+        }
         right_wnd_edge = pcb->snd_wnd + pcb->snd_wl2;
 
         /* Update window. */
@@ -1136,8 +1117,8 @@ static void tcp_receive(struct tcp_pcb *pcb, tcp_in_data *in_data)
             /* Reset the retransmission time-out. */
             pcb->rto = (pcb->sa >> 3) + pcb->sv;
 
-            /* Update the send buffer space. Diff between the two can never exceed 64K? */
-            pcb->acked = (u32_t)(in_data->ackno - pcb->lastack);
+            /* Update the send buffer space.*/
+            pcb->acked = in_data->ackno - pcb->lastack;
 
             pcb->snd_buf += pcb->acked;
 
@@ -1145,8 +1126,7 @@ static void tcp_receive(struct tcp_pcb *pcb, tcp_in_data *in_data)
             pcb->dupacks = 0;
             pcb->lastack = in_data->ackno;
 
-            /* Update the congestion control variables (cwnd and
-               ssthresh). */
+            /* Update the congestion control variables (cwnd and ssthresh). */
             if (get_tcp_state(pcb) >= ESTABLISHED) {
 #if TCP_CC_ALGO_MOD
                 cc_ack_received(pcb, CC_ACK);
@@ -1175,52 +1155,12 @@ static void tcp_receive(struct tcp_pcb *pcb, tcp_in_data *in_data)
                      ? ntohl(pcb->unacked->tcphdr->seqno) + TCP_SEGLEN(pcb->unacked)
                      : 0));
 
-            /* Remove segment from the unacknowledged list if the incoming
-               ACK acknowlegdes them. */
-            while (pcb->unacked != NULL) {
-
-                /* The purpose of this processing is to avoid to send again
-                 * data from TSO segment that is partially acknowledged.
-                 * This TSO segment was not released in tcp_receive() because
-                 * input data processing releases whole acknowledged segment only.
-                 */
-                if (pcb->unacked->flags & TF_SEG_OPTS_TSO) {
-                    u32_t removed;
-                    removed = pcb->unacked->flags & TF_SEG_OPTS_ZEROCOPY
-                        ? tcp_shrink_zc_segment(pcb, pcb->unacked, in_data->ackno)
-                        : tcp_shrink_segment(pcb, pcb->unacked, in_data->ackno);
-                    pcb->snd_queuelen -= removed;
-                }
-
-                if (!(TCP_SEQ_LEQ(pcb->unacked->seqno + TCP_SEGLEN(pcb->unacked),
-                                  in_data->ackno))) {
-                    break;
-                }
-                LWIP_DEBUGF(TCP_INPUT_DEBUG,
-                            ("tcp_receive: removing %" U32_F ":%" U32_F " from pcb->unacked\n",
-                             ntohl(pcb->unacked->tcphdr->seqno),
-                             ntohl(pcb->unacked->tcphdr->seqno) + TCP_SEGLEN(pcb->unacked)));
-
-                next = pcb->unacked;
-                pcb->unacked = pcb->unacked->next;
-                LWIP_DEBUGF(TCP_QLEN_DEBUG,
-                            ("tcp_receive: queuelen %" U32_F " ... ", (u32_t)pcb->snd_queuelen));
-                LWIP_ASSERT("pcb->snd_queuelen >= pbuf_clen(next->p)",
-                            (pcb->snd_queuelen >= pbuf_clen(next->p)));
-                /* Prevent ACK for FIN to generate a sent event */
-                if ((pcb->acked != 0) && ((next->tcp_flags & TCP_FIN) != 0)) {
-                    pcb->acked--;
-                }
-
-                pcb->snd_queuelen -= pbuf_clen(next->p);
-                tcp_tx_seg_free(pcb, next);
-                LWIP_DEBUGF(TCP_QLEN_DEBUG,
-                            ("%" U32_F " (after freeing unacked)\n", (u32_t)pcb->snd_queuelen));
-            }
+            ack_partial_or_whole_segment(pcb, in_data->ackno, &(pcb->unacked));
 
             /* If there's nothing left to acknowledge, stop the retransmit
                timer, otherwise reset it to start again */
             if (pcb->unacked == NULL) {
+                pcb->last_unacked = NULL;
                 if (persist) {
                     /* start persist timer */
                     pcb->persist_cnt = 0;
@@ -1244,40 +1184,11 @@ static void tcp_receive(struct tcp_pcb *pcb, tcp_in_data *in_data)
            rationale is that lwIP puts all outstanding segments on the
            ->unsent list after a retransmission, so these segments may
            in fact have been sent once. */
-        while (pcb->unsent != NULL &&
-               TCP_SEQ_BETWEEN(in_data->ackno,
-                               ntohl(pcb->unsent->tcphdr->seqno) + TCP_SEGLEN(pcb->unsent),
-                               pcb->snd_nxt)) {
-            LWIP_DEBUGF(TCP_INPUT_DEBUG,
-                        ("tcp_receive: removing %" U32_F ":%" U32_F " from pcb->unsent\n",
-                         ntohl(pcb->unsent->tcphdr->seqno),
-                         ntohl(pcb->unsent->tcphdr->seqno) + TCP_SEGLEN(pcb->unsent)));
+        ack_partial_or_whole_segment(pcb, in_data->ackno, &(pcb->unsent));
 
-            next = pcb->unsent;
-            pcb->unsent = pcb->unsent->next;
-            LWIP_DEBUGF(TCP_QLEN_DEBUG,
-                        ("tcp_receive: queuelen %" U32_F " ... ", (u32_t)pcb->snd_queuelen));
-            LWIP_ASSERT("pcb->snd_queuelen >= pbuf_clen(next->p)",
-                        (pcb->snd_queuelen >= pbuf_clen(next->p)));
-            /* Prevent ACK for FIN to generate a sent event */
-            if ((pcb->acked != 0) && ((next->tcp_flags & TCP_FIN) != 0)) {
-                pcb->acked--;
-            }
-            pcb->snd_queuelen -= pbuf_clen(next->p);
-            tcp_tx_seg_free(pcb, next);
-            LWIP_DEBUGF(TCP_QLEN_DEBUG,
-                        ("%" U16_F " (after freeing unsent)\n", (u32_t)pcb->snd_queuelen));
-            if (pcb->snd_queuelen != 0) {
-                LWIP_ASSERT("tcp_receive: valid queue length",
-                            pcb->unacked != NULL || pcb->unsent != NULL);
-            }
-        }
         if (pcb->unsent == NULL) {
             /* We have sent all pending segments, reflect it in last_unsent */
             pcb->last_unsent = NULL;
-#if TCP_OVERSIZE
-            pcb->unsent_oversize = 0;
-#endif /* TCP_OVERSIZE */
         }
         /* End of ACK for new data processing. */
 
@@ -1375,7 +1286,7 @@ static void tcp_receive(struct tcp_pcb *pcb, tcp_in_data *in_data)
                adjust the ->data pointer in the seg and the segment
                length.*/
 
-            off = pcb->rcv_nxt - in_data->seqno;
+            u32_t off = pcb->rcv_nxt - in_data->seqno;
             p = in_data->inseg.p;
             LWIP_ASSERT("inseg.p != NULL", in_data->inseg.p);
             if (in_data->inseg.p->len < off) {
