@@ -1,5 +1,7 @@
 /*
- * Copyright (c) 2001-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
+ * Copyright (c) 2001-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-License-Identifier: GPL-2.0-only or BSD-3-Clause
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -58,11 +60,10 @@ dst_entry::dst_entry(const sock_addr &dst, uint16_t src_port, socket_data &sock_
                                                        : (header *)(new header_ipv4()))
     , m_bound_ip(in6addr_any)
     , m_so_bindtodevice_ip(in6addr_any)
-    , m_route_src_ip(in6addr_any)
     , m_pkt_src_ip(in6addr_any)
-    , m_ring_alloc_logic_tx(sock_data.fd, ring_alloc_logic, this)
-    , m_p_tx_mem_buf_desc_list(NULL)
-    , m_p_zc_mem_buf_desc_list(NULL)
+    , m_ring_alloc_logic_tx(sock_data.fd, ring_alloc_logic)
+    , m_p_tx_mem_buf_desc_list(nullptr)
+    , m_p_zc_mem_buf_desc_list(nullptr)
     , m_b_tx_mem_buf_desc_list_pending(false)
     , m_ttl_hop_limit(sock_data.ttl_hop_limit)
     , m_tos(sock_data.tos)
@@ -91,47 +92,47 @@ dst_entry::~dst_entry()
 
     if (m_p_rt_entry) {
         g_p_route_table_mgr->unregister_observer(
-            route_rule_table_key(m_dst_ip, m_route_src_ip, m_family, m_tos), this);
-        m_p_rt_entry = NULL;
+            route_rule_table_key(m_dst_ip, m_bound_ip, m_family, m_tos), this);
+        m_p_rt_entry = nullptr;
     }
 
     if (m_p_ring) {
         if (m_sge) {
             delete[] m_sge;
-            m_sge = NULL;
+            m_sge = nullptr;
         }
 
         if (m_p_tx_mem_buf_desc_list) {
             m_p_ring->mem_buf_tx_release(m_p_tx_mem_buf_desc_list, true);
-            m_p_tx_mem_buf_desc_list = NULL;
+            m_p_tx_mem_buf_desc_list = nullptr;
         }
         if (m_p_zc_mem_buf_desc_list) {
             m_p_ring->mem_buf_tx_release(m_p_zc_mem_buf_desc_list, true);
-            m_p_zc_mem_buf_desc_list = NULL;
+            m_p_zc_mem_buf_desc_list = nullptr;
         }
 
         m_p_net_dev_val->release_ring(m_ring_alloc_logic_tx.get_key());
-        m_p_ring = NULL;
+        m_p_ring = nullptr;
     }
 
     if (m_p_send_wqe_handler) {
         delete m_p_send_wqe_handler;
-        m_p_send_wqe_handler = NULL;
+        m_p_send_wqe_handler = nullptr;
     }
 
     if (m_p_neigh_val) {
         delete m_p_neigh_val;
-        m_p_neigh_val = NULL;
+        m_p_neigh_val = nullptr;
     }
 
     if (m_header) {
         delete m_header;
-        m_header = NULL;
+        m_header = nullptr;
     }
 
     if (m_header_neigh) {
         delete m_header_neigh;
-        m_header_neigh = NULL;
+        m_header_neigh = nullptr;
     }
 
     dst_logdbg("Done %s", to_str().c_str());
@@ -140,21 +141,20 @@ dst_entry::~dst_entry()
 void dst_entry::init_members()
 {
     set_state(false);
-    m_p_rt_val = NULL;
-    m_p_net_dev_val = NULL;
-    m_p_ring = NULL;
-    m_p_net_dev_entry = NULL;
-    m_p_neigh_entry = NULL;
-    m_p_neigh_val = NULL;
-    m_p_rt_entry = NULL;
+    m_p_rt_val = nullptr;
+    m_p_net_dev_val = nullptr;
+    m_p_ring = nullptr;
+    m_p_net_dev_entry = nullptr;
+    m_p_neigh_entry = nullptr;
+    m_p_neigh_val = nullptr;
+    m_p_rt_entry = nullptr;
     memset(&m_inline_send_wqe, 0, sizeof(m_inline_send_wqe));
     memset(&m_not_inline_send_wqe, 0, sizeof(m_not_inline_send_wqe));
     memset(&m_fragmented_send_wqe, 0, sizeof(m_not_inline_send_wqe));
-    m_p_send_wqe_handler = NULL;
-    m_sge = NULL;
+    m_p_send_wqe_handler = nullptr;
+    m_sge = nullptr;
     m_b_is_offloaded = true;
     m_b_is_initialized = false;
-    m_p_send_wqe = NULL;
     m_max_inline = 0;
     m_max_ip_payload_size = 0;
     m_max_udp_payload_size = 0;
@@ -164,8 +164,8 @@ void dst_entry::init_members()
 
 void dst_entry::set_src_addr()
 {
-    if (!m_route_src_ip.is_anyaddr()) {
-        m_pkt_src_ip = m_route_src_ip;
+    if (!m_bound_ip.is_anyaddr()) {
+        m_pkt_src_ip = m_bound_ip;
         dst_logfunc("Selected source address (bind): %s",
                     m_pkt_src_ip.to_str(get_sa_family()).c_str());
     } else if (get_routing_addr_sel_src(m_pkt_src_ip)) {
@@ -192,9 +192,7 @@ void dst_entry::set_src_addr()
 bool dst_entry::get_routing_addr_sel_src(ip_address &out_ip) const
 {
     if (m_p_rt_val) {
-        // For AF_INET we keep the legacy source selection logic which uses modified routing src
-        out_ip =
-            (m_family == AF_INET ? m_p_rt_val->get_src_addr() : m_p_rt_val->get_cfg_src_addr());
+        out_ip = m_p_rt_val->get_src_addr();
         return !out_ip.is_anyaddr();
     }
 
@@ -228,7 +226,7 @@ bool dst_entry::update_net_dev_val()
             }
             g_p_neigh_table_mgr->unregister_observer(
                 neigh_key(ip_addr(dst_addr, m_family), m_p_net_dev_val), this);
-            m_p_neigh_entry = NULL;
+            m_p_neigh_entry = nullptr;
         }
 
         // Change the net_device, clean old resources...
@@ -259,7 +257,7 @@ bool dst_entry::update_net_dev_val()
 bool dst_entry::update_rt_val()
 {
     bool ret_val = true;
-    route_val *p_rt_val = NULL;
+    route_val *p_rt_val = nullptr;
 
     if (m_p_rt_entry && m_p_rt_entry->get_val(p_rt_val)) {
         if (m_p_rt_val == p_rt_val) {
@@ -276,11 +274,11 @@ bool dst_entry::update_rt_val()
     return ret_val;
 }
 
-bool dst_entry::resolve_net_dev(bool is_connect)
+bool dst_entry::resolve_net_dev()
 {
     bool ret_val = false;
 
-    cache_entry_subject<route_rule_table_key, route_val *> *p_ces = NULL;
+    cache_entry_subject<route_rule_table_key, route_val *> *p_ces = nullptr;
 
     if (m_dst_ip.is_anyaddr()) {
         dst_logdbg(PRODUCT_NAME " does not offload zero net IP address");
@@ -295,37 +293,12 @@ bool dst_entry::resolve_net_dev(bool is_connect)
     // When XLIO will support routing with OIF, we need to check changing in outgoing interface
     // Source address changes is not checked since multiple bind is not allowed on the same socket
     if (!m_p_rt_entry) {
-        m_route_src_ip = m_bound_ip;
-        route_rule_table_key rtk(m_dst_ip, m_route_src_ip, m_family, m_tos);
-        dst_logfunc("Fetching rt_entry %s", m_route_src_ip.to_str(m_family).c_str());
+        route_rule_table_key rtk(m_dst_ip, m_bound_ip, m_family, m_tos);
+        dst_logfunc("Fetching rt_entry %s", m_bound_ip.to_str(m_family).c_str());
         if (g_p_route_table_mgr->register_observer(rtk, this, &p_ces)) {
             // In case this is the first time we trying to resolve route entry,
             // means that register_observer was run
             m_p_rt_entry = dynamic_cast<route_entry *>(p_ces);
-
-            // Routing entry src-addr is used for src-addr selection algorithm and not
-            // as a key for matching routing rules. Moreover, this can be a forcefully
-            // set src addr by XLIO. We keep this logic for IPv4 only for backward compliancy.
-            if (m_family == AF_INET && is_connect && m_route_src_ip.is_anyaddr()) {
-                dst_logfunc("Checking rt_entry src addr");
-                route_val *p_rt_val = NULL;
-                if (m_p_rt_entry && m_p_rt_entry->get_val(p_rt_val) &&
-                    !p_rt_val->get_src_addr().is_anyaddr()) {
-                    g_p_route_table_mgr->unregister_observer(rtk, this);
-                    m_route_src_ip = p_rt_val->get_src_addr();
-
-                    dst_logfunc("Chaning m_route_src_ip to %s",
-                                m_route_src_ip.to_str(m_family).c_str());
-
-                    route_rule_table_key new_rtk(m_dst_ip, m_route_src_ip, m_family, m_tos);
-                    if (g_p_route_table_mgr->register_observer(new_rtk, this, &p_ces)) {
-                        m_p_rt_entry = dynamic_cast<route_entry *>(p_ces);
-                    } else {
-                        dst_logdbg("Error in route resolving logic");
-                        return ret_val;
-                    }
-                }
-            }
         } else {
             dst_logdbg("Error in registering route entry");
             return ret_val;
@@ -347,11 +320,11 @@ bool dst_entry::resolve_neigh()
     if (m_p_rt_val && !m_p_rt_val->get_gw_addr().is_anyaddr() && !dst_addr.is_mc(m_family)) {
         dst_addr = m_p_rt_val->get_gw_addr();
     }
-    cache_entry_subject<neigh_key, neigh_val *> *p_ces = NULL;
+    cache_entry_subject<neigh_key, neigh_val *> *p_ces = nullptr;
     if (m_p_neigh_entry ||
         g_p_neigh_table_mgr->register_observer(
             neigh_key(ip_addr(dst_addr, m_family), m_p_net_dev_val), this, &p_ces)) {
-        if (m_p_neigh_entry == NULL) {
+        if (!m_p_neigh_entry) {
             m_p_neigh_entry = dynamic_cast<neigh_entry *>(p_ces);
         }
         if (m_p_neigh_entry) {
@@ -379,7 +352,7 @@ bool dst_entry::resolve_ring()
         if (m_p_ring) {
             if (m_sge) {
                 delete[] m_sge;
-                m_sge = NULL;
+                m_sge = nullptr;
             }
             m_sge = new (std::nothrow) struct ibv_sge[m_p_ring->get_max_send_sge()];
             if (!m_sge) {
@@ -401,15 +374,15 @@ bool dst_entry::release_ring()
         if (m_p_ring) {
             if (m_p_tx_mem_buf_desc_list) {
                 m_p_ring->mem_buf_tx_release(m_p_tx_mem_buf_desc_list, true);
-                m_p_tx_mem_buf_desc_list = NULL;
+                m_p_tx_mem_buf_desc_list = nullptr;
             }
             if (m_p_zc_mem_buf_desc_list) {
                 m_p_ring->mem_buf_tx_release(m_p_zc_mem_buf_desc_list, true);
-                m_p_zc_mem_buf_desc_list = NULL;
+                m_p_zc_mem_buf_desc_list = nullptr;
             }
             dst_logdbg("releasing a ring");
             m_p_net_dev_val->release_ring(m_ring_alloc_logic_tx.get_key());
-            m_p_ring = NULL;
+            m_p_ring = nullptr;
         }
         ret_val = true;
     }
@@ -456,7 +429,7 @@ bool dst_entry::conf_l2_hdr_and_snd_wqe_eth()
     // scratch
     if (m_p_send_wqe_handler) {
         delete m_p_send_wqe_handler;
-        m_p_send_wqe_handler = NULL;
+        m_p_send_wqe_handler = nullptr;
     }
 
     m_p_send_wqe_handler = new wqe_send_handler();
@@ -540,8 +513,7 @@ bool dst_entry::offloaded_according_to_rules()
     return ret_val;
 }
 
-bool dst_entry::prepare_to_send(struct xlio_rate_limit_t &rate_limit, bool skip_rules,
-                                bool is_connect)
+bool dst_entry::prepare_to_send(struct xlio_rate_limit_t &rate_limit, bool skip_rules)
 {
     bool resolved = false;
     m_slow_path_lock.lock();
@@ -557,7 +529,7 @@ bool dst_entry::prepare_to_send(struct xlio_rate_limit_t &rate_limit, bool skip_
     if (!m_b_force_os && !is_valid()) {
         bool b_is_offloaded = false;
         set_state(true);
-        if (resolve_net_dev(is_connect)) {
+        if (resolve_net_dev()) {
             set_src_addr();
             // overwrite mtu from route if exists
             m_max_udp_payload_size =
@@ -584,11 +556,11 @@ bool dst_entry::prepare_to_send(struct xlio_rate_limit_t &rate_limit, bool skip_
                                                  m_src_port, m_dst_port);
                     if (m_p_tx_mem_buf_desc_list) {
                         m_p_ring->mem_buf_tx_release(m_p_tx_mem_buf_desc_list, true);
-                        m_p_tx_mem_buf_desc_list = NULL;
+                        m_p_tx_mem_buf_desc_list = nullptr;
                     }
                     if (m_p_zc_mem_buf_desc_list) {
                         m_p_ring->mem_buf_tx_release(m_p_zc_mem_buf_desc_list, true);
-                        m_p_zc_mem_buf_desc_list = NULL;
+                        m_p_zc_mem_buf_desc_list = nullptr;
                     }
                     resolved = true;
                 }
@@ -681,7 +653,7 @@ void dst_entry::do_ring_migration_tx(lock_base &socket_lock, resource_allocation
     m_p_ring = new_ring;
     if (m_sge) {
         delete[] m_sge;
-        m_sge = NULL;
+        m_sge = nullptr;
     }
     m_sge = new (std::nothrow) struct ibv_sge[m_p_ring->get_max_send_sge()];
     if (!m_sge) {
@@ -692,9 +664,9 @@ void dst_entry::do_ring_migration_tx(lock_base &socket_lock, resource_allocation
                                       get_route_mtu() + (uint32_t)m_header->m_transport_header_len);
 
     mem_buf_desc_t *tmp_list = m_p_tx_mem_buf_desc_list;
-    m_p_tx_mem_buf_desc_list = NULL;
+    m_p_tx_mem_buf_desc_list = nullptr;
     mem_buf_desc_t *tmp_list_zc = m_p_zc_mem_buf_desc_list;
-    m_p_zc_mem_buf_desc_list = NULL;
+    m_p_zc_mem_buf_desc_list = nullptr;
 
     m_slow_path_lock.unlock();
     socket_lock.unlock();
@@ -735,7 +707,7 @@ uint16_t dst_entry::get_dst_port()
     return m_dst_port;
 }
 
-ssize_t dst_entry::pass_buff_to_neigh(const iovec *p_iov, size_t sz_iov, uint32_t packet_id)
+ssize_t dst_entry::pass_pkt_to_neigh(const iovec *p_iov, size_t sz_iov, uint32_t packet_id)
 {
     ssize_t ret_val = 0;
 
@@ -771,7 +743,7 @@ bool dst_entry::alloc_neigh_val(transport_type_t tranport)
 
     if (m_p_neigh_val) {
         delete m_p_neigh_val;
-        m_p_neigh_val = NULL;
+        m_p_neigh_val = nullptr;
     }
 
     switch (tranport) {
@@ -790,25 +762,25 @@ void dst_entry::return_buffers_pool()
 {
     int count;
 
-    if (m_p_tx_mem_buf_desc_list == NULL && m_p_zc_mem_buf_desc_list == NULL) {
+    if (!m_p_tx_mem_buf_desc_list && !m_p_zc_mem_buf_desc_list) {
         return;
     }
 
     if (m_b_tx_mem_buf_desc_list_pending && m_p_ring) {
-        if (m_p_tx_mem_buf_desc_list != NULL) {
+        if (m_p_tx_mem_buf_desc_list) {
             count = m_p_ring->mem_buf_tx_release(m_p_tx_mem_buf_desc_list, true, true);
             if (count) {
-                m_p_tx_mem_buf_desc_list = NULL;
+                m_p_tx_mem_buf_desc_list = nullptr;
             }
         }
-        if (m_p_zc_mem_buf_desc_list != NULL) {
+        if (m_p_zc_mem_buf_desc_list) {
             count = m_p_ring->mem_buf_tx_release(m_p_zc_mem_buf_desc_list, true, true);
             if (count) {
-                m_p_zc_mem_buf_desc_list = NULL;
+                m_p_zc_mem_buf_desc_list = nullptr;
             }
         }
     }
-    set_tx_buff_list_pending(m_p_tx_mem_buf_desc_list != NULL || m_p_zc_mem_buf_desc_list != NULL);
+    set_tx_buff_list_pending(m_p_tx_mem_buf_desc_list || m_p_zc_mem_buf_desc_list);
 }
 
 int dst_entry::modify_ratelimit(struct xlio_rate_limit_t &rate_limit)
@@ -833,7 +805,7 @@ bool dst_entry::update_ring_alloc_logic(int fd, lock_base &socket_lock,
 {
     resource_allocation_key old_key(*m_ring_alloc_logic_tx.get_key());
 
-    m_ring_alloc_logic_tx = ring_allocation_logic_tx(fd, ring_alloc_logic, this);
+    m_ring_alloc_logic_tx = ring_allocation_logic_tx(fd, ring_alloc_logic);
 
     if (*m_ring_alloc_logic_tx.get_key() != old_key) {
         std::lock_guard<decltype(m_tx_migration_lock)> locker(m_tx_migration_lock);
