@@ -1,5 +1,7 @@
 /*
- * Copyright (c) 2001-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
+ * Copyright (c) 2001-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-License-Identifier: GPL-2.0-only or BSD-3-Clause
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -35,12 +37,11 @@
 
 #include "utils/lock_wrapper.h"
 #include "proto/mem_buf_desc.h"
-#include "sock/socket_fd_api.h"
+#include "sock/sockinfo.h"
 #include "dev/buffer_pool.h"
-#include "dev/cq_mgr.h"
+#include "dev/cq_mgr_rx.h"
 #include "xlio_extra.h"
 
-// LWIP includes
 #include "lwip/opt.h"
 #include "lwip/tcp_impl.h"
 
@@ -48,6 +49,10 @@
 #include "sockinfo_ulp.h"
 #include "sockinfo_nvme.h"
 
+/* Forward declarations */
+struct xlio_socket_attr;
+class poll_group;
+
 #define BLOCK_THIS_RUN(blocking, flags) (blocking && !(flags & MSG_DONTWAIT))
 
 /**
@@ -89,6 +94,14 @@ enum tcp_conn_state_e {
     TCP_CONN_RESETED
 };
 
+enum xlio_express_flags : uint32_t {
+    XLIO_EXPRESS_OP_TYPE_DESC,
+    XLIO_EXPRESS_OP_TYPE_FILE_ZEROCOPY,
+    XLIO_EXPRESS_OP_TYPE_MASK = 0x000fu,
+    XLIO_EXPRESS_MSG_MORE,
+    XLIO_EXPRESS_MSG_MASK = 0x00f0u,
+};
+
 struct socket_option_t {
     const int level;
     const int optname;
@@ -113,6 +126,51 @@ struct socket_option_t {
     }
 };
 
+class tcp_timers_collection : public timer_handler, public cleanable_obj {
+public:
+    tcp_timers_collection();
+    tcp_timers_collection(int intervals);
+    ~tcp_timers_collection() override;
+
+    void clean_obj() override;
+
+    void handle_timer_expired(void *user_data) override;
+
+    void register_wakeup_event();
+
+    void add_new_timer(sockinfo_tcp *sock);
+
+    void remove_timer(sockinfo_tcp *sock);
+
+    void set_group(poll_group *group) { m_p_group = group; }
+    inline event_handler_manager *get_event_mgr();
+
+private:
+    void free_tta_resources();
+
+protected:
+    void *m_timer_handle = nullptr;
+
+private:
+    typedef std::list<sockinfo_tcp *> sock_list;
+    typedef typename sock_list::iterator sock_list_itr;
+    std::vector<sock_list> m_p_intervals;
+    std::unordered_map<sockinfo_tcp *, std::tuple<uint32_t, sock_list_itr>> m_sock_remove_map;
+    int m_n_intervals_size;
+    int m_n_location = 0;
+    int m_n_count = 0;
+    int m_n_next_insert_bucket = 0;
+    poll_group *m_p_group = nullptr;
+};
+
+class thread_local_tcp_timers : public tcp_timers_collection {
+public:
+    thread_local_tcp_timers();
+    ~thread_local_tcp_timers() override;
+};
+
+extern tcp_timers_collection *g_tcp_timers_collection;
+
 typedef std::deque<socket_option_t *> socket_options_list_t;
 typedef std::map<tcp_pcb *, int> ready_pcb_map_t;
 typedef std::map<flow_tuple, tcp_pcb *> syn_received_map_t;
@@ -127,83 +185,90 @@ enum inet_ecns {
     INET_ECN_MASK = 3,
 };
 
-class sockinfo_tcp : public sockinfo, public timer_handler {
+class sockinfo_tcp : public sockinfo {
 public:
-    static inline size_t accepted_conns_node_offset(void)
+    static inline size_t accepted_conns_node_offset()
     {
         return NODE_OFFSET(sockinfo_tcp, accepted_conns_node);
     }
     typedef xlio_list_t<sockinfo_tcp, sockinfo_tcp::accepted_conns_node_offset> sock_list_t;
     sockinfo_tcp(int fd, int domain);
-    virtual ~sockinfo_tcp();
+    ~sockinfo_tcp() override;
 
-    virtual void clean_obj();
+    void clean_socket_obj() override;
 
     void setPassthrough(bool _isPassthrough)
     {
         m_sock_offload = _isPassthrough ? TCP_SOCK_PASSTHROUGH : TCP_SOCK_LWIP;
-        m_p_socket_stats->b_is_offloaded = !_isPassthrough;
+        IF_STATS(m_p_socket_stats->b_is_offloaded = !_isPassthrough);
     }
-    void setPassthrough() { setPassthrough(true); }
-    bool isPassthrough() { return m_sock_offload == TCP_SOCK_PASSTHROUGH; }
+    void setPassthrough() override { setPassthrough(true); }
+    bool isPassthrough() override { return m_sock_offload == TCP_SOCK_PASSTHROUGH; }
 
-    int prepareListen();
-    int shutdown(int __how);
+    int prepareListen() override;
+    int shutdown(int __how) override;
 
     // Not always we can close immediately TCP socket: we can do that only after the TCP connection
     // in closed. In this method we just kikstarting the TCP connection termination (empty the
     // unsent/unacked, senf FIN...) Return val: true is the socket is already closable and false
     // otherwise
-    virtual bool prepare_to_close(bool process_shutdown = false);
+    bool prepare_to_close(bool process_shutdown = false) override;
     void create_dst_entry();
+    void destructor_helper_tcp();
     bool prepare_dst_to_send(bool is_accepted_socket = false);
 
-    virtual int fcntl(int __cmd, unsigned long int __arg);
-    virtual int fcntl64(int __cmd, unsigned long int __arg);
-    virtual int ioctl(unsigned long int __request, unsigned long int __arg);
-    virtual int setsockopt(int __level, int __optname, const void *__optval, socklen_t __optlen);
+    int fcntl(int __cmd, unsigned long int __arg) override;
+    int fcntl64(int __cmd, unsigned long int __arg) override;
+    int ioctl(unsigned long int __request, unsigned long int __arg) override;
+    int setsockopt(int __level, int __optname, const void *__optval, socklen_t __optlen) override;
     virtual int tcp_setsockopt(int __level, int __optname, const void *__optval,
                                socklen_t __optlen);
-    virtual int getsockopt(int __level, int __optname, void *__optval, socklen_t *__optlen);
+    int getsockopt(int __level, int __optname, void *__optval, socklen_t *__optlen) override;
     int getsockopt_offload(int __level, int __optname, void *__optval, socklen_t *__optlen);
-    virtual int connect(const sockaddr *, socklen_t);
-    virtual int bind(const sockaddr *__addr, socklen_t __addrlen);
-    virtual int listen(int backlog);
-    virtual int accept(struct sockaddr *__addr, socklen_t *__addrlen);
-    virtual int accept4(struct sockaddr *__addr, socklen_t *__addrlen, int __flags);
-    virtual int getsockname(sockaddr *__name, socklen_t *__namelen);
-    virtual int getpeername(sockaddr *__name, socklen_t *__namelen);
+    int connect(const sockaddr *, socklen_t) override;
+    int bind(const sockaddr *__addr, socklen_t __addrlen) override;
+    int listen(int backlog) override;
+    int accept(struct sockaddr *__addr, socklen_t *__addrlen) override;
+    int accept4(struct sockaddr *__addr, socklen_t *__addrlen, int __flags) override;
+    int getsockname(sockaddr *__name, socklen_t *__namelen) override;
+    int getpeername(sockaddr *__name, socklen_t *__namelen) override;
+    void set_immediate_os_sample() override {};
+    void unset_immediate_os_sample() override {};
 
     inline bool handle_bind_no_port(int &bind_ret, in_port_t in_port, const sockaddr *__addr,
                                     socklen_t __addrlen);
     inline void non_tcp_recved(int rx_len);
-    virtual int recvfrom_zcopy_free_packets(struct xlio_recvfrom_zcopy_packet_t *pkts,
-                                            size_t count);
+    int recvfrom_zcopy_free_packets(struct xlio_recvfrom_zcopy_packet_t *pkts,
+                                    size_t count) override;
 
     void socketxtreme_recv_buffs_tcp(mem_buf_desc_t *desc, uint16_t len);
 
-    virtual void statistics_print(vlog_levels_t log_level = VLOG_DEBUG);
+    void statistics_print(vlog_levels_t log_level = VLOG_DEBUG) override;
 
-    inline struct tcp_pcb *get_pcb(void) { return &m_pcb; }
+    inline struct tcp_pcb *get_pcb() { return &m_pcb; }
 
-    inline unsigned sndbuf_available(void) { return tcp_sndbuf(&m_pcb); }
+    inline unsigned sndbuf_available()
+    {
+        return static_cast<unsigned>(std::max(tcp_sndbuf(&m_pcb), 0));
+    }
 
-    inline unsigned get_mss(void) { return m_pcb.mss; }
+    inline unsigned get_mss() { return m_pcb.mss; }
 
-    ssize_t tx(xlio_tx_call_attr_t &tx_arg);
+    ssize_t tx(xlio_tx_call_attr_t &tx_arg) override;
     ssize_t tcp_tx(xlio_tx_call_attr_t &tx_arg);
     ssize_t rx(const rx_call_t call_type, iovec *p_iov, ssize_t sz_iov, int *p_flags,
-               sockaddr *__from = NULL, socklen_t *__fromlen = NULL, struct msghdr *__msg = NULL);
+               sockaddr *__from = nullptr, socklen_t *__fromlen = nullptr,
+               struct msghdr *__msg = nullptr) override;
     static err_t ip_output(struct pbuf *p, struct tcp_seg *seg, void *v_p_conn, uint16_t flags);
     static err_t ip_output_syn_ack(struct pbuf *p, struct tcp_seg *seg, void *v_p_conn,
                                    uint16_t flags);
     static void tcp_state_observer(void *pcb_container, enum tcp_state new_state);
     static uint16_t get_route_mtu(struct tcp_pcb *pcb);
 
-    virtual void update_header_field(data_updater *updater);
-    virtual bool rx_input_cb(mem_buf_desc_t *p_rx_pkt_mem_buf_desc_info, void *pv_fd_ready_array);
+    void update_header_field(data_updater *updater) override;
+    bool rx_input_cb(mem_buf_desc_t *p_rx_pkt_mem_buf_desc_info, void *pv_fd_ready_array) override;
     void abort_connection();
-    void tcp_shutdown_rx(void);
+    void tcp_shutdown_rx();
 
     mem_buf_desc_t *tcp_tx_mem_buf_alloc(pbuf_type type);
     void tcp_rx_mem_buf_free(mem_buf_desc_t *p_desc);
@@ -215,22 +280,23 @@ public:
     static struct tcp_seg *tcp_seg_alloc_cached(void *p_conn);
     static void tcp_seg_free_direct(void *p_conn, struct tcp_seg *seg);
     static void tcp_seg_free_cached(void *p_conn, struct tcp_seg *seg);
-    uint32_t get_next_tcp_seqno(void) { return m_pcb.snd_lbb; }
-    uint32_t get_next_tcp_seqno_rx(void) { return m_pcb.rcv_nxt; }
+    uint32_t get_next_tcp_seqno() { return m_pcb.snd_lbb; }
+    uint32_t get_next_tcp_seqno_rx() { return m_pcb.rcv_nxt; }
 
     mem_buf_desc_t *tcp_tx_zc_alloc(mem_buf_desc_t *p_desc);
+    static void tcp_express_zc_callback(mem_buf_desc_t *p_desc);
     static void tcp_tx_zc_callback(mem_buf_desc_t *p_desc);
     void tcp_tx_zc_handle(mem_buf_desc_t *p_desc);
 
-    bool inline is_readable(uint64_t *p_poll_sn, fd_array_t *p_fd_array = NULL);
-    bool inline is_writeable();
-    bool inline is_errorable(int *errors);
-    bool is_closable()
+    bool is_readable(uint64_t *p_poll_sn, fd_array_t *p_fd_array = NULL) override;
+    bool is_writeable() override;
+    bool is_errorable(int *errors) override;
+    bool is_closable() override
     {
         return get_tcp_state(&m_pcb) == CLOSED && m_syn_received.empty() &&
             m_accepted_conns.empty();
     }
-    bool inline is_destroyable_lock(void)
+    bool inline is_destroyable_lock()
     {
         bool state;
         m_tcp_con_lock.lock();
@@ -238,25 +304,27 @@ public:
         m_tcp_con_lock.unlock();
         return state;
     }
-    bool inline is_destroyable_no_lock(void)
+    bool inline is_destroyable_no_lock()
     {
         return get_tcp_state(&m_pcb) == CLOSED && m_state == SOCKINFO_CLOSING;
     }
-    bool skip_os_select()
+    bool skip_os_select() override
     {
         // calling os select on offloaded TCP sockets makes no sense unless it's a listen socket
         // to make things worse, it returns that os fd is ready...
         return (m_sock_offload == TCP_SOCK_LWIP && !is_server() && m_conn_state != TCP_CONN_INIT);
     }
 
-    bool is_outgoing()
+    bool is_outgoing() override
     {
         const bool is_listen_socket = is_server() || get_tcp_state(&m_pcb) == LISTEN;
         // Excluding incoming and listen sockets we can determine outgoing sockets.
         return !m_b_incoming && !is_listen_socket;
     }
 
-    bool is_incoming() { return m_b_incoming; }
+    bool is_incoming() override { return m_b_incoming; }
+    bool is_timer_registered() const { return m_timer_registered; }
+    void set_timer_registered(bool v) { m_timer_registered = v; }
 
     bool is_connected() { return m_sock_state == TCP_SOCK_CONNECTED_RDWR; }
 
@@ -277,33 +345,41 @@ public:
         return m_sock_state == TCP_SOCK_ACCEPT_READY || m_sock_state == TCP_SOCK_ACCEPT_SHUT;
     }
 
-    virtual void update_socket_timestamps(timestamps_t *ts) { m_rx_timestamps = *ts; }
+    void update_socket_timestamps(timestamps_t *ts) override { m_rx_timestamps = *ts; }
 
-    static const int CONNECT_DEFAULT_TIMEOUT_MS = 10000;
-    virtual inline fd_type_t get_type() { return FD_TYPE_SOCKET; }
+    inline fd_type_t get_type() override { return FD_TYPE_SOCKET; }
 
-    void handle_timer_expired(void *user_data);
+    void handle_timer_expired();
 
-    inline ib_ctx_handler *get_ctx(void)
+    inline ib_ctx_handler *get_ctx()
     {
         return m_p_connected_dst_entry ? m_p_connected_dst_entry->get_ctx() : nullptr;
     }
 
-    inline ring *get_tx_ring(void) const noexcept
+    inline ring *get_tx_ring() const noexcept
     {
         return m_p_connected_dst_entry ? m_p_connected_dst_entry->get_ring() : nullptr;
     }
 
-    inline ring *get_rx_ring(void) { return m_p_rx_ring; }
-    const flow_tuple_with_local_if &get_flow_tuple(void)
+    void rx_add_ring_cb(ring *p_ring) override;
+    ring *get_rx_ring() { return m_p_rx_ring; }
+    const flow_tuple_with_local_if &get_flow_tuple()
     {
-        /* XXX Dosn't handle empty map and a map with multiple elements. */
+        /* XXX Doesn't handle empty map and a map with multiple elements. */
         auto rx_flow_iter = m_rx_flow_map.begin();
         return rx_flow_iter->first;
     }
 
+    void rx_poll_on_tx_if_needed()
+    {
+        if (m_sysvar_rx_poll_on_tx_tcp) {
+            int poll_count = 0;
+            rx_wait_helper(poll_count, false);
+        }
+    }
+
     /* Proxy to support ULP. TODO Refactor. */
-    inline sockinfo_tcp_ops *get_ops(void) { return m_ops; }
+    inline sockinfo_tcp_ops *get_ops() { return m_ops; }
     inline void set_ops(sockinfo_tcp_ops *ops) noexcept
     {
         std::swap(ops, m_ops);
@@ -311,107 +387,57 @@ public:
             delete ops;
         }
     }
-    inline void reset_ops(void) noexcept { set_ops(m_ops_tcp); }
+    inline void reset_ops() noexcept { set_ops(m_ops_tcp); }
 
     bool is_utls_supported(int direction) const;
 
     int get_supported_nvme_feature_mask() const;
 
-    inline int trylock_tcp_con(void) { return m_tcp_con_lock.trylock(); }
-    inline void lock_tcp_con(void) { m_tcp_con_lock.lock(); }
-    inline void unlock_tcp_con(void) { m_tcp_con_lock.unlock(); }
+    inline int trylock_tcp_con() { return m_tcp_con_lock.trylock(); }
+    inline void lock_tcp_con() { m_tcp_con_lock.lock(); }
+    inline void unlock_tcp_con() { m_tcp_con_lock.unlock(); }
+    tcp_timers_collection *get_tcp_timer_collection();
+    bool is_cleaned() const { return m_is_cleaned; }
+    static err_t rx_lwip_cb(void *arg, struct tcp_pcb *tpcb, struct pbuf *p, err_t err);
+    static err_t rx_lwip_cb_socketxtreme(void *arg, struct tcp_pcb *tpcb, struct pbuf *p,
+                                         err_t err);
+    static err_t rx_lwip_cb_recv_callback(void *arg, struct tcp_pcb *pcb, struct pbuf *p,
+                                          err_t err);
+    static err_t rx_drop_lwip_cb(void *arg, struct tcp_pcb *tpcb, struct pbuf *p, err_t err);
+    inline void rx_lwip_cb_socketxtreme_helper(pbuf *p);
 
-    list_node<sockinfo_tcp, sockinfo_tcp::accepted_conns_node_offset> accepted_conns_node;
+    int register_callback(xlio_recv_callback_t callback, void *context) override
+    {
+        tcp_recv(&m_pcb, sockinfo_tcp::rx_lwip_cb_recv_callback);
+        return register_callback_ctx(callback, context);
+    }
 
-    inline void set_reguired_send_block(unsigned sz) { m_required_send_block = sz; }
+    int tcp_tx_express(const struct iovec *iov, unsigned iov_len, uint32_t mkey, unsigned flags,
+                       void *opaque_op);
+    int tcp_tx_express_inline(const struct iovec *iov, unsigned iov_len, unsigned flags);
+    void flush();
+
+    void set_xlio_socket(const struct xlio_socket_attr *attr);
+    void add_tx_ring_to_group();
+    bool is_xlio_socket() { return m_p_group != nullptr; }
+    poll_group *get_poll_group() { return m_p_group; }
+    void xlio_socket_event(int event, int value);
+    static err_t rx_lwip_cb_xlio_socket(void *arg, struct tcp_pcb *tpcb, struct pbuf *p, err_t err);
+    static void err_lwip_cb_xlio_socket(void *pcb_container, err_t err);
 
 protected:
-    virtual void lock_rx_q();
-    virtual void unlock_rx_q();
-    virtual bool try_un_offloading(); // un-offload the socket if possible
-    virtual int os_epoll_wait(epoll_event *ep_events, int maxevents);
+    void lock_rx_q() override;
+    void unlock_rx_q() override;
+    bool try_un_offloading() override; // un-offload the socket if possible
+    int os_epoll_wait(epoll_event *ep_events, int maxevents) override;
+
+    size_t handle_msg_trunc(size_t total_rx, size_t payload_size, int in_flags,
+                            int *p_out_flags) override;
 
 private:
     int fcntl_helper(int __cmd, unsigned long int __arg, bool &bexit);
     void get_tcp_info(struct tcp_info *ti);
 
-    sockinfo_tcp_ops *m_ops;
-    sockinfo_tcp_ops *m_ops_tcp;
-
-    // lwip specific things
-    struct tcp_pcb m_pcb;
-    socket_options_list_t m_socket_options_list;
-    timestamps_t m_rx_timestamps;
-    tcp_sock_offload_e m_sock_offload;
-    tcp_sock_state_e m_sock_state;
-    sockinfo_tcp *m_parent;
-    // received packet source (true if its from internal thread)
-    bool m_xlio_thr;
-    bool m_b_incoming;
-    bool m_b_attached;
-    /* connection state machine */
-    int m_conn_timeout;
-    /* SNDBUF acconting */
-    int m_sndbuff_max;
-    /* RCVBUF acconting */
-    int m_rcvbuff_max;
-    int m_rcvbuff_current;
-    int m_rcvbuff_non_tcp_recved;
-    tcp_conn_state_e m_conn_state;
-    fd_array_t *m_iomux_ready_fd_array;
-    struct linger m_linger;
-
-    /* local & peer addresses */
-    /*	struct sockaddr *m_addr_local;
-        socklen_t m_local_alen;
-        struct sockaddr *m_addr_peer;
-        socklen_t m_peer_alen;
-    */
-
-    // Relevant only for listen sockets: map connections in syn received state
-    // We need this map since for syn received connection no sockinfo is created yet!
-    syn_received_map_t m_syn_received;
-    uint32_t m_received_syn_num;
-
-    /* pending connections */
-    sock_list_t m_accepted_conns;
-
-    uint32_t m_ready_conn_cnt;
-    int m_backlog;
-
-    void *m_timer_handle;
-    multilock m_tcp_con_lock;
-
-    // used for reporting 'connected' on second non-blocking call to connect or
-    // second call to failed connect blocking socket.
-    bool report_connected;
-
-    int m_error_status;
-
-    const buffer_batching_mode_t m_sysvar_buffer_batching_mode;
-    const uint32_t m_sysvar_tx_segs_batch_tcp;
-    const option_tcp_ctl_thread::mode_t m_sysvar_tcp_ctl_thread;
-
-    struct tcp_seg *m_tcp_seg_list;
-    uint32_t m_tcp_seg_count;
-    uint32_t m_tcp_seg_in_use;
-
-    xlio_desc_list_t m_rx_pkt_ready_list;
-    xlio_desc_list_t m_rx_cb_dropped_list;
-
-    lock_spin_recursive m_rx_ctl_packets_list_lock;
-    tscval_t m_last_syn_tsc;
-    xlio_desc_list_t m_rx_ctl_packets_list;
-    peer_map_t m_rx_peer_packets;
-    xlio_desc_list_t m_rx_ctl_reuse_list;
-    ready_pcb_map_t m_ready_pcbs;
-    static const unsigned TX_CONSECUTIVE_EAGAIN_THREASHOLD = 10;
-    unsigned m_tx_consecutive_eagain_count;
-    bool m_sysvar_rx_poll_on_tx_tcp;
-    uint64_t m_user_huge_page_mask;
-    unsigned m_required_send_block;
-    uint16_t m_external_vlan_tag = 0U;
-
     inline void lwip_pbuf_init_custom(mem_buf_desc_t *p_desc);
 
     void tcp_timer();
@@ -448,33 +474,26 @@ private:
     int wait_for_conn_ready_blocking();
     static err_t connect_lwip_cb(void *arg, struct tcp_pcb *tpcb, err_t err);
     // tx
-    unsigned tx_wait(int &err, bool blocking);
+    unsigned tx_wait(bool blocking);
     int os_epoll_wait_with_tcp_timers(epoll_event *ep_events, int maxevents);
-    int handle_child_FIN(sockinfo_tcp *child_conn);
+    void handle_incoming_handshake_failure(sockinfo_tcp *child_conn);
 
     // rx
     // int rx_wait(int &poll_count, bool blocking = true);
     static err_t ack_recvd_lwip_cb(void *arg, struct tcp_pcb *tpcb, u16_t space);
 
-public:
-    static err_t rx_lwip_cb(void *arg, struct tcp_pcb *tpcb, struct pbuf *p, err_t err);
-    static err_t rx_lwip_cb_socketxtreme(void *arg, struct tcp_pcb *tpcb, struct pbuf *p,
-                                         err_t err);
-    static err_t rx_lwip_cb_recv_callback(void *arg, struct tcp_pcb *pcb, struct pbuf *p,
-                                          err_t err);
-    static err_t rx_drop_lwip_cb(void *arg, struct tcp_pcb *tpcb, struct pbuf *p, err_t err);
-    inline void rx_lwip_cb_socketxtreme_helper(pbuf *p);
-
-    virtual int register_callback(xlio_recv_callback_t callback, void *context)
-    {
-        tcp_recv(&m_pcb, sockinfo_tcp::rx_lwip_cb_recv_callback);
-        return sockinfo::register_callback(callback, context);
-    }
-
-private:
-    inline err_t handle_fin(struct tcp_pcb *pcb, err_t err);
-    inline void handle_rx_lwip_cb_error(pbuf *p);
-    inline void rx_lwip_cb_error(pbuf *p);
+    ssize_t tcp_tx_handle_done_and_unlock(ssize_t total_tx, int errno_tmp, bool is_dummy,
+                                          bool is_send_zerocopy);
+    ssize_t tcp_tx_handle_errno_and_unlock(int error_number);
+    ssize_t tcp_tx_handle_partial_send_and_unlock(ssize_t total_tx, int errno_to_report,
+                                                  bool is_dummy, bool is_send_zerocopy,
+                                                  int errno_to_restore);
+    ssize_t tcp_tx_handle_sndbuf_unavailable(ssize_t total_tx, bool is_dummy, bool is_send_zerocopy,
+                                             int errno_to_restore);
+    ssize_t tcp_tx_slow_path(xlio_tx_call_attr_t &tx_arg);
+    err_t handle_fin(struct tcp_pcb *pcb, err_t err);
+    void handle_rx_lwip_cb_error(pbuf *p);
+    void rx_lwip_cb_error(pbuf *p);
     inline void rx_lwip_process_chained_pbufs(pbuf *p);
     inline void rx_lwip_shrink_rcv_wnd(size_t pbuf_tot_len, int nbytes);
     inline void save_packet_info_in_ready_list(pbuf *p);
@@ -490,7 +509,7 @@ private:
     /*
      * Supported only for UDP
      */
-    virtual void handle_ip_pktinfo(struct cmsg_state *) {};
+    void handle_ip_pktinfo(struct cmsg_state *) override {};
 
     int handle_rx_error(bool blocking);
 
@@ -511,9 +530,9 @@ private:
     inline void return_pending_rx_buffs();
     inline void return_pending_tx_buffs();
     inline void reuse_buffer(mem_buf_desc_t *buff);
-    virtual mem_buf_desc_t *get_next_desc(mem_buf_desc_t *p_desc);
-    virtual mem_buf_desc_t *get_next_desc_peek(mem_buf_desc_t *p_desc, int &rx_pkt_ready_list_idx);
-    virtual timestamps_t *get_socket_timestamps();
+    mem_buf_desc_t *get_next_desc(mem_buf_desc_t *p_desc) override;
+    mem_buf_desc_t *get_next_desc_peek(mem_buf_desc_t *p_desc, int &rx_pkt_ready_list_idx) override;
+    timestamps_t *get_socket_timestamps() override;
 
     inline void return_reuse_buffers_postponed()
     {
@@ -527,7 +546,7 @@ private:
         m_rx_reuse_buf_postponed = false;
 
         if (m_p_rx_ring) {
-            if (m_rx_reuse_buff.n_buff_num >= m_n_sysvar_rx_num_buffs_reuse) {
+            if (m_rx_reuse_buff.n_buff_num >= m_rx_num_buffs_reuse) {
                 if (m_p_rx_ring->reclaim_recv_buffers(&m_rx_reuse_buff.rx_reuse)) {
                     m_rx_reuse_buff.n_buff_num = 0;
                 } else {
@@ -539,7 +558,7 @@ private:
             while (iter != m_rx_ring_map.end()) {
                 descq_t *rx_reuse = &iter->second->rx_reuse_info.rx_reuse;
                 int &n_buff_num = iter->second->rx_reuse_info.n_buff_num;
-                if (n_buff_num >= m_n_sysvar_rx_num_buffs_reuse) {
+                if (n_buff_num >= m_rx_num_buffs_reuse) {
                     if (iter->first->reclaim_recv_buffers(rx_reuse)) {
                         n_buff_num = 0;
                     } else {
@@ -551,32 +570,27 @@ private:
         }
     }
 
-    virtual void post_deqeue(bool release_buff);
-    virtual int zero_copy_rx(iovec *p_iov, mem_buf_desc_t *pdesc, int *p_flags);
+    void post_deqeue(bool release_buff) override;
+    int zero_copy_rx(iovec *p_iov, mem_buf_desc_t *pdesc, int *p_flags) override;
 
     // Returns the connected pcb, with 5 tuple which matches the input arguments,
     // in state "SYN Received" or NULL if pcb wasn't found
     struct tcp_pcb *get_syn_received_pcb(const flow_tuple &key) const;
     struct tcp_pcb *get_syn_received_pcb(const sock_addr &src, const sock_addr &dst);
 
-    virtual mem_buf_desc_t *get_front_m_rx_pkt_ready_list();
-    virtual size_t get_size_m_rx_pkt_ready_list();
-    virtual void pop_front_m_rx_pkt_ready_list();
-    virtual void push_back_m_rx_pkt_ready_list(mem_buf_desc_t *buff);
-
-    // stats
-    uint64_t m_n_pbufs_rcvd;
-    uint64_t m_n_pbufs_freed;
+    mem_buf_desc_t *get_front_m_rx_pkt_ready_list() override;
+    size_t get_size_m_rx_pkt_ready_list() override;
+    void pop_front_m_rx_pkt_ready_list() override;
+    void push_back_m_rx_pkt_ready_list(mem_buf_desc_t *buff) override;
 
     // lock_spin_recursive m_rx_cq_lck;
     /* pick all cqs that match given address */
-    virtual int rx_verify_available_data();
+    int rx_verify_available_data() override;
     inline int rx_wait(int &poll_count, bool blocking);
     inline int rx_wait_lockless(int &poll_count, bool blocking);
     int rx_wait_helper(int &poll_count, bool blocking);
     void fit_rcv_wnd(bool force_fit);
     void fit_snd_bufs(unsigned int new_max);
-    void fit_snd_bufs_to_nagle(bool disable_nagle);
 
     inline struct tcp_seg *get_tcp_seg_cached();
     inline struct tcp_seg *get_tcp_seg_direct();
@@ -591,48 +605,94 @@ private:
     void process_reuse_ctl_packets();
     void process_rx_ctl_packets();
     static void put_agent_msg(void *arg);
-};
-typedef struct tcp_seg tcp_seg;
+    bool is_connected_and_ready_to_send();
+
+    inline event_handler_manager *get_event_mgr();
 
-class tcp_timers_collection : public timers_group, public cleanable_obj {
 public:
-    tcp_timers_collection(int period, int resolution);
-    virtual ~tcp_timers_collection();
+    static const int CONNECT_DEFAULT_TIMEOUT_MS = 10000;
 
-    void clean_obj();
-
-    virtual void handle_timer_expired(void *user_data);
-
-protected:
-    // add a new timer
-    void add_new_timer(timer_node_t *node, timer_handler *handler, void *user_data);
-
-    // remove timer from list and free it.
-    // called for stopping (unregistering) a timer
-    void remove_timer(timer_node_t *node);
-
-    void *m_timer_handle;
+    list_node<sockinfo_tcp, sockinfo_tcp::accepted_conns_node_offset> accepted_conns_node;
 
 private:
-    timer_node_t **m_p_intervals;
+    sockinfo_tcp_ops *m_ops;
+    sockinfo_tcp_ops *m_ops_tcp;
 
-    int m_n_period;
-    int m_n_resolution;
-    int m_n_intervals_size;
-    int m_n_location;
-    int m_n_count;
-    int m_n_next_insert_bucket;
+    // lwip specific things
+    struct tcp_pcb m_pcb;
+    fd_array_t *m_iomux_ready_fd_array;
+    socket_options_list_t m_socket_options_list;
+    timestamps_t m_rx_timestamps;
+    tcp_sock_offload_e m_sock_offload;
+    tcp_sock_state_e m_sock_state;
+    sockinfo_tcp *m_parent;
+    // received packet source (true if its from internal thread)
+    bool m_xlio_thr;
+    bool m_b_incoming;
+    bool m_b_attached;
+    bool m_timer_registered = false;
+    /* connection state machine */
+    int m_conn_timeout;
+    /* RCVBUF acconting */
+    int m_rcvbuff_max;
+    int m_rcvbuff_current;
+    int m_rcvbuff_non_tcp_recved;
+    tcp_conn_state_e m_conn_state;
+    struct linger m_linger;
 
-    void free_tta_resources();
+    /* local & peer addresses */
+    /*	struct sockaddr *m_addr_local;
+        socklen_t m_local_alen;
+        struct sockaddr *m_addr_peer;
+        socklen_t m_peer_alen;
+    */
+
+    // Relevant only for listen sockets: map connections in syn received state
+    // We need this map since for syn received connection no sockinfo is created yet!
+    syn_received_map_t m_syn_received;
+
+    /* pending connections */
+    sock_list_t m_accepted_conns;
+
+    uint32_t m_ready_conn_cnt;
+    int m_backlog;
+
+    multilock m_tcp_con_lock;
+
+    // used for reporting 'connected' on second non-blocking call to connect or
+    // second call to failed connect blocking socket.
+    bool report_connected;
+    bool m_is_cleaned = false; // If this socket registered deletion on internal thread.
+    int m_error_status;
+
+    const buffer_batching_mode_t m_sysvar_buffer_batching_mode;
+    const uint32_t m_sysvar_tx_segs_batch_tcp;
+
+    struct tcp_seg *m_tcp_seg_list;
+    uint32_t m_tcp_seg_count;
+    uint32_t m_tcp_seg_in_use;
+
+    xlio_desc_list_t m_rx_pkt_ready_list;
+    xlio_desc_list_t m_rx_cb_dropped_list;
+
+    lock_spin_recursive m_rx_ctl_packets_list_lock;
+    tscval_t m_last_syn_tsc;
+    xlio_desc_list_t m_rx_ctl_packets_list;
+    peer_map_t m_rx_peer_packets;
+    xlio_desc_list_t m_rx_ctl_reuse_list;
+    ready_pcb_map_t m_ready_pcbs;
+    static const unsigned TX_CONSECUTIVE_EAGAIN_THREASHOLD = 10;
+    unsigned m_tx_consecutive_eagain_count;
+    bool m_sysvar_rx_poll_on_tx_tcp;
+    uint64_t m_user_huge_page_mask;
+    uint16_t m_external_vlan_tag = 0U;
+    /*
+     * Storage API
+     * TODO Move the fields to proper cold/hot sections in the final version.
+     */
+    bool m_b_xlio_socket_dirty = false;
+    uintptr_t m_xlio_socket_userdata = 0;
+    poll_group *m_p_group = nullptr;
 };
 
-class thread_local_tcp_timers : public tcp_timers_collection {
-public:
-    thread_local_tcp_timers();
-    ~thread_local_tcp_timers();
-};
-
-extern tcp_timers_collection *g_tcp_timers_collection;
-extern thread_local thread_local_tcp_timers g_thread_local_tcp_timers;
-
 #endif
