@@ -19,27 +19,6 @@
 
 #define UCP_PROTO_RNDV_PUT_DESC "write to remote"
 
-
-enum {
-    /* Initial stage for put zcopy is sending the data */
-    UCP_PROTO_RNDV_PUT_ZCOPY_STAGE_SEND = UCP_PROTO_STAGE_START,
-
-    /* Initial stage for put memtype is copy the data to the fragment */
-    UCP_PROTO_RNDV_PUT_MTYPE_STAGE_COPY = UCP_PROTO_STAGE_START,
-
-    /* Flush all lanes to ensure remote delivery */
-    UCP_PROTO_RNDV_PUT_STAGE_FLUSH,
-
-    /* Send ATP without fence (could be done after a flush) */
-    UCP_PROTO_RNDV_PUT_STAGE_ATP,
-
-    /* Send ATP with fence (could be done if using send lanes for ATP) */
-    UCP_PROTO_RNDV_PUT_STAGE_FENCED_ATP,
-
-    /* Memtype only: send the fragment to the remote side */
-    UCP_PROTO_RNDV_PUT_MTYPE_STAGE_SEND
-};
-
 typedef struct ucp_proto_rndv_put_priv {
     uct_completion_callback_t  put_comp_cb;
     uct_completion_callback_t  atp_comp_cb;
@@ -47,6 +26,7 @@ typedef struct ucp_proto_rndv_put_priv {
     ucp_lane_map_t             flush_map;
     ucp_lane_map_t             atp_map;
     ucp_lane_index_t           atp_num_lanes;
+    uint8_t                    stat_counter;
     ucp_proto_rndv_bulk_priv_t bulk;
 } ucp_proto_rndv_put_priv_t;
 
@@ -61,12 +41,22 @@ typedef struct {
 static UCS_F_ALWAYS_INLINE void
 ucp_proto_rndv_put_common_complete(ucp_request_t *req)
 {
+    const ucp_proto_rndv_put_priv_t UCS_V_UNUSED *rpriv =
+                                                   req->send.proto_config->priv;
     ucp_trace_req(req, "rndv_put_common_complete");
-    UCP_WORKER_STAT_RNDV(req->send.ep->worker, PUT_ZCOPY, +1);
+    UCS_STATS_UPDATE_COUNTER(req->send.ep->worker->stats, rpriv->stat_counter,
+                             +1);
     ucp_proto_rndv_rkey_destroy(req);
     ucp_proto_request_zcopy_complete(req, req->send.state.uct_comp.status);
 }
 
+static void ucp_proto_rndv_put_zcopy_completion(uct_completion_t *uct_comp)
+{
+    ucp_request_t *req = ucs_container_of(uct_comp, ucp_request_t,
+                                          send.state.uct_comp);
+    ucp_proto_rndv_put_common_complete(req);
+}
+
 static UCS_F_ALWAYS_INLINE ucs_status_t
 ucp_proto_rndv_put_common_send(ucp_request_t *req,
                                const ucp_proto_multi_lane_priv_t *lpriv,
@@ -116,9 +106,11 @@ static ucs_status_t
 ucp_proto_rndv_put_common_flush_progress(uct_pending_req_t *uct_req)
 {
     ucp_request_t *req = ucs_container_of(uct_req, ucp_request_t, send.uct);
+    const ucp_proto_rndv_put_priv_t *rpriv;
 
+    rpriv = req->send.proto_config->priv;
     return ucp_proto_multi_lane_map_progress(
-            req, &req->send.rndv.put.flush_map,
+            req, &req->send.rndv.put.flush_lane, rpriv->flush_map,
             ucp_proto_rndv_put_common_flush_send);
 }
 
@@ -134,35 +126,59 @@ ucp_proto_rndv_put_common_atp_send(ucp_request_t *req, ucp_lane_index_t lane)
 {
     const ucp_proto_rndv_put_priv_t *rpriv = req->send.proto_config->priv;
     ucp_proto_rndv_put_atp_pack_ctx_t pack_ctx;
+    ucs_status_t status;
+
+    /* Make sure the sum of ack_size field in all ATP messages we send will not
+       exceed request length, since each ATP message has to acknowledge at least
+       one byte. */
+    ucs_assertv(req->send.rndv.put.atp_count <= req->send.state.dt_iter.length,
+                "atp_count=%u length=%zu", req->send.rndv.put.atp_count,
+                req->send.state.dt_iter.length);
+    if (req->send.rndv.put.atp_count == req->send.state.dt_iter.length) {
+        return UCS_OK;
+    }
+
+    /* Ensure ATP is sent on the same lane as the data to prevent ATP from
+     * arriving before the data. If data transmission starts from a non-zero
+     * lane, ATP may never be sent on the data lane. */
+    if (ucs_unlikely((req->send.state.dt_iter.length < rpriv->atp_num_lanes) &&
+                     (lane < req->send.multi_lane_idx))) {
+        return UCS_OK;
+    }
 
     pack_ctx.req = req;
 
     /* When we need to send multiple ATP messages: each will acknowledge 1 byte,
        except the last ATP which will acknowledge the remaining payload size.
        This is simpler than keeping track of how much was sent on each lane */
-    ucs_assert(req->send.rndv.put.atp_map != 0);
-    if (ucs_is_pow2(req->send.rndv.put.atp_map)) {
+    if (req->send.rndv.put.atp_count == (rpriv->atp_num_lanes - 1)) {
         pack_ctx.ack_size = req->send.state.dt_iter.length -
-                            rpriv->atp_num_lanes + 1;
-        if (pack_ctx.ack_size == 0) {
-            return UCS_OK; /* Skip sending 0-length ATP */
-        }
+                            req->send.rndv.put.atp_count;
     } else {
         pack_ctx.ack_size = 1;
     }
 
-    return ucp_proto_am_bcopy_single_send(req, UCP_AM_ID_RNDV_ATP, lane,
-                                          ucp_proto_rndv_put_common_pack_atp,
-                                          &pack_ctx, sizeof(ucp_rndv_ack_hdr_t),
-                                          0);
+    status = ucp_proto_am_bcopy_single_send(req, UCP_AM_ID_RNDV_ATP, lane,
+                                            ucp_proto_rndv_put_common_pack_atp,
+                                            &pack_ctx,
+                                            sizeof(ucp_rndv_ack_hdr_t), 0);
+    if (status != UCS_OK) {
+        return status;
+    }
+
+    ++req->send.rndv.put.atp_count;
+    return UCS_OK;
 }
 
 static ucs_status_t
 ucp_proto_rndv_put_common_atp_progress(uct_pending_req_t *uct_req)
 {
     ucp_request_t *req = ucs_container_of(uct_req, ucp_request_t, send.uct);
+    const ucp_proto_rndv_put_priv_t *rpriv;
 
-    return ucp_proto_multi_lane_map_progress(req, &req->send.rndv.put.atp_map,
+    rpriv = req->send.proto_config->priv;
+    return ucp_proto_multi_lane_map_progress(req, &req->send.rndv.put.atp_lane,
+                                             rpriv->atp_map,
                                              ucp_proto_rndv_put_common_atp_send);
 }
 
@@ -184,9 +200,11 @@ static ucs_status_t
 ucp_proto_rndv_put_common_fenced_atp_progress(uct_pending_req_t *uct_req)
 {
     ucp_request_t *req = ucs_container_of(uct_req, ucp_request_t, send.uct);
+    const ucp_proto_rndv_put_priv_t *rpriv;
 
+    rpriv = req->send.proto_config->priv;
     return ucp_proto_multi_lane_map_progress(
-            req, &req->send.rndv.put.atp_map,
+            req, &req->send.rndv.put.atp_lane, rpriv->atp_map,
             ucp_proto_rndv_put_common_fenced_atp_send);
 }
 
@@ -205,28 +223,29 @@ ucp_proto_rndv_put_common_request_init(ucp_request_t *req)
 {
     const ucp_proto_rndv_put_priv_t *rpriv = req->send.proto_config->priv;
 
-    req->send.rndv.put.atp_map   = rpriv->atp_map;
-    req->send.rndv.put.flush_map = rpriv->flush_map;
+    req->send.rndv.put.flush_lane = 0;
+    req->send.rndv.put.atp_lane   = 0;
+    req->send.rndv.put.atp_count  = 0;
     ucp_proto_rndv_bulk_request_init(req, &rpriv->bulk);
 }
 
-static ucs_status_t
-ucp_proto_rndv_put_common_init(const ucp_proto_init_params_t *init_params,
-                               uint64_t rndv_modes, size_t max_length,
-                               uct_ep_operation_t memtype_op, unsigned flags,
-                               ucp_md_map_t initial_reg_md_map,
-                               uct_completion_callback_t comp_cb,
-                               int support_ppln)
+static void
+ucp_proto_rndv_put_common_probe(const ucp_proto_init_params_t *init_params,
+                                uint64_t rndv_modes, size_t max_length,
+                                uct_ep_operation_t memtype_op, unsigned flags,
+                                ucp_md_map_t initial_reg_md_map,
+                                uct_completion_callback_t comp_cb,
+                                int support_ppln, uint8_t stat_counter,
+                                const ucp_memory_info_t *reg_mem_info)
 {
     const size_t atp_size                = sizeof(ucp_rndv_ack_hdr_t);
     ucp_context_t *context               = init_params->worker->context;
-    ucp_proto_rndv_put_priv_t *rpriv     = init_params->priv;
     ucp_proto_multi_init_params_t params = {
         .super.super         = *init_params,
         .super.overhead      = 0,
         .super.latency       = 0,
         .super.cfg_thresh    = ucp_proto_rndv_cfg_thresh(context, rndv_modes),
-        .super.cfg_priority  = 0,
+        .super.cfg_priority  = 80,
         .super.min_length    = 0,
         .super.max_length    = max_length,
         .super.min_iov       = 1,
@@ -238,8 +257,10 @@ ucp_proto_rndv_put_common_init(const ucp_proto_init_params_t *init_params,
         .super.send_op       = UCT_EP_OP_PUT_ZCOPY,
         .super.memtype_op    = memtype_op,
         .super.flags         = flags | UCP_PROTO_COMMON_INIT_FLAG_RECV_ZCOPY |
-                               UCP_PROTO_COMMON_INIT_FLAG_REMOTE_ACCESS,
+                               UCP_PROTO_COMMON_INIT_FLAG_REMOTE_ACCESS |
+                               UCP_PROTO_COMMON_INIT_FLAG_MIN_FRAG,
         .super.exclude_map   = 0,
+        .super.reg_mem_info  = *reg_mem_info,
         .max_lanes           = context->config.ext.max_rndv_lanes,
         .initial_reg_md_map  = initial_reg_md_map,
         .first.tl_cap_flags  = UCT_IFACE_FLAG_PUT_ZCOPY,
@@ -252,79 +273,94 @@ ucp_proto_rndv_put_common_init(const ucp_proto_init_params_t *init_params,
     };
     const uct_iface_attr_t *iface_attr;
     ucp_lane_index_t lane_idx, lane;
+    ucp_proto_rndv_put_priv_t rpriv;
     int send_atp, use_fence;
-    size_t bulk_priv_size;
+    ucp_proto_perf_t *perf;
     ucs_status_t status;
+    unsigned atp_map;
 
     if ((init_params->select_param->dt_class != UCP_DATATYPE_CONTIG) ||
         !ucp_proto_rndv_op_check(init_params, UCP_OP_ID_RNDV_SEND,
                                  support_ppln) ||
         !ucp_proto_common_init_check_err_handling(&params.super)) {
-        return UCS_ERR_UNSUPPORTED;
+        return;
     }
 
-    status = ucp_proto_rndv_bulk_init(&params, &rpriv->bulk,
-                                      UCP_PROTO_RNDV_PUT_DESC,
-                                      UCP_PROTO_RNDV_ATP_NAME, &bulk_priv_size);
+    status = ucp_proto_rndv_bulk_init(&params, UCP_PROTO_RNDV_PUT_DESC,
+                                      UCP_PROTO_RNDV_ATP_NAME, &perf,
+                                      &rpriv.bulk);
     if (status != UCS_OK) {
-        return status;
+        return;
     }
 
-    *init_params->priv_size = ucs_offsetof(ucp_proto_rndv_put_priv_t, bulk) +
-                              bulk_priv_size;
+    send_atp = !ucp_proto_rndv_init_params_is_ppln_frag(init_params);
 
-    /* Check if all potential lanes support sending ATP */
-    rpriv     = params.super.super.priv;
-    send_atp  = !ucp_proto_rndv_init_params_is_ppln_frag(init_params);
-    use_fence = send_atp && !context->config.ext.rndv_put_force_flush;
-
-    /* Check if all potential lanes support sending ATP */
-    lane_idx  = 0;
-    while (use_fence && (lane_idx < rpriv->bulk.mpriv.num_lanes)) {
-        lane       = rpriv->bulk.mpriv.lanes[lane_idx++].super.lane;
+    /* Check which lanes support sending ATP */
+    atp_map = 0;
+    for (lane_idx = 0; lane_idx < rpriv.bulk.mpriv.num_lanes; ++lane_idx) {
+        lane       = rpriv.bulk.mpriv.lanes[lane_idx].super.lane;
         iface_attr = ucp_proto_common_get_iface_attr(init_params, lane);
-        use_fence  = use_fence &&
-                     (((iface_attr->cap.flags & UCT_IFACE_FLAG_AM_SHORT) &&
-                       (iface_attr->cap.am.max_short >= atp_size)) ||
-                      ((iface_attr->cap.flags & UCT_IFACE_FLAG_AM_BCOPY) &&
-                       (iface_attr->cap.am.max_bcopy >= atp_size)));
+        if (((iface_attr->cap.flags & UCT_IFACE_FLAG_AM_SHORT) &&
+             (iface_attr->cap.am.max_short >= atp_size)) ||
+            ((iface_attr->cap.flags & UCT_IFACE_FLAG_AM_BCOPY) &&
+             (iface_attr->cap.am.max_bcopy >= atp_size))) {
+            atp_map |= UCS_BIT(lane);
+        }
     }
 
+    /* Use fence only if all lanes support sending ATP and flush is not forced
+     */
+    use_fence = send_atp && !context->config.ext.rndv_put_force_flush &&
+                (rpriv.bulk.mpriv.lane_map == atp_map);
+
     /* All lanes can send ATP - invalidate am_lane, to use mpriv->lanes.
-     * Otherwise, would need to flush all lanes and send ATP on
-     * rpriv->super.lane when the flush is completed
+     * Otherwise, would need to flush all lanes and send ATP on:
+     * - All lanes supporting ATP send. This ensures that data is flushed
+     *   remotely (i.e. resides in the target buffer), which may not be the case
+     *   with IB transports. An alternative would be to pass
+     *   UCT_FLUSH_FLAG_REMOTE to uct_ep_flush, but using this flag increases
+     *   UCP worker address size.
+     *   TODO: Consider calling UCT ep flush with remote flag when/if address
+     *   size is not an issue anymore.
+     * - Control lane if none of the lanes support sending ATP
      */
     if (use_fence) {
         /* Send fence followed by ATP on all lanes */
-        rpriv->bulk.super.lane = UCP_NULL_LANE;
-        rpriv->put_comp_cb     = comp_cb;
-        rpriv->atp_comp_cb     = NULL;
-        rpriv->stage_after_put = UCP_PROTO_RNDV_PUT_STAGE_FENCED_ATP;
-        rpriv->flush_map       = 0;
-        rpriv->atp_map         = rpriv->bulk.mpriv.lane_map;
+        rpriv.bulk.super.lane = UCP_NULL_LANE;
+        rpriv.put_comp_cb     = comp_cb;
+        rpriv.atp_comp_cb     = NULL;
+        rpriv.stage_after_put = UCP_PROTO_RNDV_PUT_STAGE_FENCED_ATP;
+        rpriv.flush_map       = 0;
+        rpriv.atp_map         = rpriv.bulk.mpriv.lane_map;
     } else {
-        /* Flush all lanes and send single ATP on control message lane */
+        /* Flush all lanes and send ATP on all supporting lanes (or control lane
+         * otherwise) */
         if (send_atp) {
-            rpriv->put_comp_cb =
+            rpriv.put_comp_cb =
                     ucp_proto_rndv_put_common_flush_completion_send_atp;
-            rpriv->atp_comp_cb = comp_cb;
-            rpriv->atp_map     = UCS_BIT(rpriv->bulk.super.lane);
+            rpriv.atp_comp_cb = comp_cb;
+            rpriv.atp_map     = (atp_map == 0) ?
+                                UCS_BIT(rpriv.bulk.super.lane) : atp_map;
         } else {
-            rpriv->put_comp_cb = comp_cb;
-            rpriv->atp_comp_cb = NULL;
-            rpriv->atp_map     = 0;
+            rpriv.put_comp_cb = comp_cb;
+            rpriv.atp_comp_cb = NULL;
+            rpriv.atp_map     = 0;
         }
-        rpriv->stage_after_put = UCP_PROTO_RNDV_PUT_STAGE_FLUSH;
-        rpriv->flush_map       = rpriv->bulk.mpriv.lane_map;
-        ucs_assert(rpriv->flush_map != 0);
+        rpriv.stage_after_put = UCP_PROTO_RNDV_PUT_STAGE_FLUSH;
+        rpriv.flush_map       = rpriv.bulk.mpriv.lane_map;
+        ucs_assert(rpriv.flush_map != 0);
     }
 
     if (send_atp) {
-        ucs_assert(rpriv->atp_map != 0);
+        ucs_assert(rpriv.atp_map != 0);
     }
-    rpriv->atp_num_lanes = ucs_popcount(rpriv->atp_map);
+    rpriv.atp_num_lanes = ucs_popcount(rpriv.atp_map);
+    rpriv.stat_counter  = stat_counter;
 
-    return UCS_OK;
+    ucp_proto_select_add_proto(&params.super.super, params.super.cfg_thresh,
+                               params.super.cfg_priority, perf, &rpriv,
+                               UCP_PROTO_MULTI_EXTENDED_PRIV_SIZE(&rpriv,
+                                                                  bulk.mpriv));
 }
 
 static const char *
@@ -383,24 +419,21 @@ ucp_proto_rndv_put_zcopy_send_progress(uct_pending_req_t *uct_req)
             ucp_proto_rndv_put_common_data_sent, rpriv->put_comp_cb);
 }
 
-static void ucp_proto_rndv_put_zcopy_completion(uct_completion_t *uct_comp)
+static void
+ucp_proto_rndv_put_zcopy_probe(const ucp_proto_init_params_t *init_params)
 {
-    ucp_request_t *req = ucs_container_of(uct_comp, ucp_request_t,
-                                          send.state.uct_comp);
-    ucp_proto_rndv_put_common_complete(req);
-}
+    ucp_memory_info_t reg_mem_info = {
+        .type    = init_params->select_param->mem_type,
+        .sys_dev = init_params->select_param->sys_dev
+    };
 
-static ucs_status_t
-ucp_proto_rndv_put_zcopy_init(const ucp_proto_init_params_t *init_params)
-{
-    unsigned flags = UCP_PROTO_COMMON_INIT_FLAG_SEND_ZCOPY |
-                     UCP_PROTO_COMMON_INIT_FLAG_ERR_HANDLING;
-
-    return ucp_proto_rndv_put_common_init(init_params,
-                                          UCS_BIT(UCP_RNDV_MODE_PUT_ZCOPY),
-                                          SIZE_MAX, UCT_EP_OP_LAST, flags, 0,
-                                          ucp_proto_rndv_put_zcopy_completion,
-                                          0);
+    ucp_proto_rndv_put_common_probe(
+            init_params, UCS_BIT(UCP_RNDV_MODE_PUT_ZCOPY), SIZE_MAX,
+            UCT_EP_OP_LAST,
+            UCP_PROTO_COMMON_INIT_FLAG_SEND_ZCOPY |
+            UCP_PROTO_COMMON_INIT_FLAG_ERR_HANDLING,
+            0, ucp_proto_rndv_put_zcopy_completion, 0,
+            UCP_WORKER_STAT_RNDV_PUT_ZCOPY, &reg_mem_info);
 }
 
 static void
@@ -414,11 +447,33 @@ ucp_proto_rndv_put_zcopy_query(const ucp_proto_query_params_t *params,
                       UCP_PROTO_ZCOPY_DESC, put_desc);
 }
 
+static ucs_status_t ucp_proto_rndv_put_zcopy_reset(ucp_request_t *req)
+{
+    const ucp_proto_rndv_put_priv_t *rpriv = req->send.proto_config->priv;
+
+    if (req->send.rndv.put.atp_count == rpriv->atp_num_lanes) {
+        /* Sent all ATPs so the iterator should be at the end */
+        ucs_assertv_always(ucp_datatype_iter_is_end(&req->send.state.dt_iter),
+                           "req=%p offset=%zu length=%zu", req,
+                           req->send.state.dt_iter.offset,
+                           req->send.state.dt_iter.length);
+    } else {
+        /* Last ATP was not sent yet or length was less than number of lanes -
+           in both cases, each sent ATP acknowledged 1 byte. */
+        ucp_datatype_iter_seek(&req->send.state.dt_iter,
+                               req->send.rndv.put.atp_count,
+                               UCS_BIT(UCP_DATATYPE_CONTIG));
+    }
+
+    req->flags &= ~UCP_REQUEST_FLAG_PROTO_INITIALIZED;
+    return UCS_OK;
+}
+
 ucp_proto_t ucp_rndv_put_zcopy_proto = {
     .name     = "rndv/put/zcopy",
     .desc     = NULL,
     .flags    = 0,
-    .init     = ucp_proto_rndv_put_zcopy_init,
+    .probe    = ucp_proto_rndv_put_zcopy_probe,
     .query    = ucp_proto_rndv_put_zcopy_query,
     .progress = {
         [UCP_PROTO_RNDV_PUT_ZCOPY_STAGE_SEND] = ucp_proto_rndv_put_zcopy_send_progress,
@@ -427,7 +482,7 @@ ucp_proto_t ucp_rndv_put_zcopy_proto = {
         [UCP_PROTO_RNDV_PUT_STAGE_FENCED_ATP] = ucp_proto_rndv_put_common_fenced_atp_progress,
     },
     .abort    = ucp_proto_request_zcopy_abort,
-    .reset    = (ucp_request_reset_func_t)ucp_proto_reset_fatal_not_implemented
+    .reset    = ucp_proto_rndv_put_zcopy_reset
 };
 
 
@@ -460,25 +515,26 @@ static UCS_F_ALWAYS_INLINE ucs_status_t ucp_proto_rndv_put_mtype_send_func(
 static ucs_status_t
 ucp_proto_rndv_put_mtype_copy_progress(uct_pending_req_t *uct_req)
 {
-    ucp_request_t *req = ucs_container_of(uct_req, ucp_request_t, send.uct);
+    ucp_request_t *req                     = ucs_container_of(uct_req,
+                                                              ucp_request_t,
+                                                              send.uct);
+    const ucp_proto_rndv_put_priv_t *rpriv = req->send.proto_config->priv;
     ucs_status_t status;
 
     ucs_assert(!(req->flags & UCP_REQUEST_FLAG_PROTO_INITIALIZED));
 
-    status = ucp_proto_rndv_mtype_request_init(req);
+    status = ucp_proto_rndv_mtype_request_init(req, rpriv->bulk.frag_mem_type);
     if (status != UCS_OK) {
         ucp_proto_request_abort(req, status);
         return UCS_OK;
     }
 
     ucp_proto_rndv_put_common_request_init(req);
-    ucp_proto_rndv_mtype_copy(req, req->send.rndv.mdesc->ptr,
-                              ucp_proto_rndv_mtype_get_req_memh(req),
-                              uct_ep_get_zcopy,
-                              ucp_proto_rndv_put_mtype_pack_completion,
-                              "in from");
-
     req->flags |= UCP_REQUEST_FLAG_PROTO_INITIALIZED;
+    ucp_proto_rndv_mdesc_mtype_copy(req, uct_ep_get_zcopy,
+                                    ucp_proto_rndv_put_mtype_pack_completion,
+                                    "in from");
+
     return UCS_OK;
 }
 
@@ -517,46 +573,73 @@ static void ucp_proto_rndv_put_mtype_frag_completion(uct_completion_t *uct_comp)
     ucp_proto_rndv_ppln_send_frag_complete(req, 1);
 }
 
-static ucs_status_t
-ucp_proto_rndv_put_mtype_init(const ucp_proto_init_params_t *init_params)
+static void
+ucp_proto_rndv_put_mtype_probe(const ucp_proto_init_params_t *init_params)
 {
+    ucp_context_t *context = init_params->worker->context;
     uct_completion_callback_t comp_cb;
     ucp_md_map_t mdesc_md_map;
     ucs_status_t status;
     size_t frag_size;
+    unsigned flags;
+    ucp_md_index_t UCS_V_UNUSED dummy_md_id;
+    ucp_memory_info_t frag_mem_info;
 
-    status = ucp_proto_rndv_mtype_init(init_params, &mdesc_md_map, &frag_size);
-    if (status != UCS_OK) {
-        return status;
+    if (init_params->rkey_config_key == NULL) {
+        return;
     }
 
+    /* Can initialize only the same fragment type as received in RTR
+     * because pipeline protocols assume that both peers use the same
+     * fragment sizes (and they are different for different memory types by
+     * default). */
+    frag_mem_info.type = init_params->rkey_config_key->mem_type;
+
+    status = ucp_proto_rndv_mtype_init(init_params, frag_mem_info.type,
+                                       &mdesc_md_map, &frag_size);
+    if (status != UCS_OK) {
+        return;
+    }
+
+    status = ucp_mm_get_alloc_md_index(context, frag_mem_info.type,
+                                       &dummy_md_id,
+                                       &frag_mem_info.sys_dev);
+    if (status != UCS_OK) {
+        return;
+    }
+
+    flags = context->config.ext.rndv_errh_ppln_enable ?
+            UCP_PROTO_COMMON_INIT_FLAG_ERR_HANDLING : 0;
+
     if (ucp_proto_rndv_init_params_is_ppln_frag(init_params)) {
         comp_cb = ucp_proto_rndv_put_mtype_frag_completion;
     } else {
         comp_cb = ucp_proto_rndv_put_mtype_completion;
     }
 
-    return ucp_proto_rndv_put_common_init(init_params,
-                                          UCS_BIT(UCP_RNDV_MODE_PUT_PIPELINE),
-                                          frag_size, UCT_EP_OP_GET_ZCOPY, 0,
-                                          mdesc_md_map, comp_cb, 1);
+    ucp_proto_rndv_put_common_probe(
+            init_params, UCS_BIT(UCP_RNDV_MODE_PUT_PIPELINE), frag_size,
+            UCT_EP_OP_GET_ZCOPY, flags, mdesc_md_map, comp_cb, 1,
+            UCP_WORKER_STAT_RNDV_PUT_MTYPE_ZCOPY, &frag_mem_info);
 }
 
 static void
 ucp_proto_rndv_put_mtype_query(const ucp_proto_query_params_t *params,
                                ucp_proto_query_attr_t *attr)
 {
+    const ucp_proto_rndv_put_priv_t *rpriv = params->priv;
     const char *put_desc;
 
     put_desc = ucp_proto_rndv_put_common_query(params, attr);
-    ucp_proto_rndv_mtype_query_desc(params, attr, put_desc);
+    ucp_proto_rndv_mtype_query_desc(params, rpriv->bulk.frag_mem_type, attr,
+                                    put_desc);
 }
 
 ucp_proto_t ucp_rndv_put_mtype_proto = {
     .name     = "rndv/put/mtype",
     .desc     = NULL,
     .flags    = 0,
-    .init     = ucp_proto_rndv_put_mtype_init,
+    .probe    = ucp_proto_rndv_put_mtype_probe,
     .query    = ucp_proto_rndv_put_mtype_query,
     .progress = {
         [UCP_PROTO_RNDV_PUT_MTYPE_STAGE_COPY] = ucp_proto_rndv_put_mtype_copy_progress,
@@ -565,6 +648,6 @@ ucp_proto_t ucp_rndv_put_mtype_proto = {
         [UCP_PROTO_RNDV_PUT_STAGE_ATP]        = ucp_proto_rndv_put_common_atp_progress,
         [UCP_PROTO_RNDV_PUT_STAGE_FENCED_ATP] = ucp_proto_rndv_put_common_fenced_atp_progress,
     },
-    .abort    = ucp_proto_abort_fatal_not_implemented,
+    .abort    = ucp_proto_rndv_stub_abort,
     .reset    = (ucp_request_reset_func_t)ucp_proto_reset_fatal_not_implemented
 };
